#!/usr/bin/env python3

import os, sys
from time import gmtime, strftime
import argparse
from pathlib import Path

if __name__ == "g_var" and 'start_dir' not in locals():
    parser = argparse.ArgumentParser(description='Converts CG representation into an atomistic representation', prog='CG2AT2', epilog='Enjoy the program and best of luck!\n')
    parser.add_argument('-info', help=' provides version, available force fields and fragments', action='store_true')
    parser.add_argument('-version', action='version', version='%(prog)s 0.2.lite')
    parser.add_argument('-c', help='coarse-grained coordinates',metavar='pdb/gro/tpr',type=str)
    parser.add_argument('-a', help='atomistic coordinates (Optional)',metavar='pdb/gro/tpr',type=str, nargs='*')
    parser.add_argument('-d', help='duplicate atomistic chains. (0:3 1:3 means 3 copies each of chain 0 and 1)',type=str, nargs='*', default=[], metavar='0:1')
    parser.add_argument('-group', help='treat user supplied atomistic chains, as rigid bodies. (0,1 2,3 or all or chain)',type=str, nargs='*', metavar='0,1')
    parser.add_argument('-loc', help='output folder name, (default = CG2AT_timestamp)',metavar='CG2AT',type=str, default='CG2AT_'+strftime("%Y-%m-%d_%H-%M-%S", gmtime()))
    parser.add_argument('-o', help='Final output supplied (default = all)', default='all', type=str, choices= ['all', 'align', 'de_novo', 'none'])
    parser.add_argument('-w', help='choose your solvent, common choices are: tip3p, tip4p, spc and spce. This is optional',metavar='tip3p',type=str)
    parser.add_argument('-ff', help='choose your force field. (Optional)',metavar='charmm36',type=str)
    parser.add_argument('-fg', help='choose your fragment library. (Optional)',metavar='martini-2-2',type=str, nargs='*')
    parser.add_argument('-mod', help='treat fragments individually', action='store_true')
    parser.add_argument('-swap', help='creates a swap dictionary supply residues as PIP2,D3A:PVCL2,C3A (Optional)',metavar='PIP2,D3A:PVCL2,C3A',type=str, nargs='*')
    parser.add_argument('-v', action="count", default=0, help="increase output verbosity (eg -vv, 3 levels) (Optional)")
    parser.add_argument('-ter', help='interactively choose terminal species (Optional)', action='store_true')
    parser.add_argument('-nt', help='choose neutral N terminal state', action='store_true')
    parser.add_argument('-ct', help='choose neutral C terminal state', action='store_true')
    parser.add_argument('-messy', help='do not remove part files CG2AT', action='store_true')
    parser.add_argument('-gmx', help='gromacs executable name (Optional)',metavar='gmx_avx',type=str)
    parser.add_argument('-cys', help='cutoff for disulphide bonds, sometimes CYS are too far apart (Optional)',metavar='7',type=float, default=7)
    parser.add_argument('-silent', help='silent cysteines question', action='store_true')
    parser.add_argument('-box', help='box size in Angstrom (0 = use input file) (Optional)',metavar='100',type=float, nargs=3)
    parser.add_argument('-vs', help='use virtual sites', action='store_true')
    parser.add_argument('-sf', help='scale factor for fragments, shrinks fragments before fitting to CG',metavar='0.9',type=float, default=0.9)
    parser.add_argument('-ncpus', help='DEPRECATED due to gromacs issues. maximum number of cores to use (default = all)', type=int)
    parser.add_argument('-disre', help='switches on the distance restraint matrix for the backbone', action='store_true')
    parser.add_argument('-ov', help='amount of overlap allowed between atoms', type=float, default=0.3, metavar='0.3')
    parser.add_argument('-posre', help='non protein residue to generate posre file', type=str, metavar='POPC')
    parser.add_argument('-compare', help='compare itp file to database', type=str, metavar='martini_v3.itp')
    args = parser.parse_args()
    opt = vars(args)
    opt['input']=os.path.abspath(sys.argv[0])+' '+''.join([ i+' ' for i in sys.argv[1:]])+'\n'

    #### virtual site information
    if args.vs:
        vs = '-vsite h'
        sf = args.sf-0.1
    else:
        vs = ''
        sf=args.sf

    ### hardcoded variables for use elsewhere in the script

    topology = {'HYDRATION':'','ALT_RES':'', 'C_TERMINAL':'default', 'N_TERMINAL':'default', 
                'CHIRAL':{'atoms':[]}, 'GROUPS':{'group_max':1}, 'CONNECT':{'atoms':{}}}

    box_line="CRYST1 %8.3f %8.3f %8.3f %8.2f %8.2f %8.2f P 1           1\n"

    pdbline = "ATOM  %5d %4s %4s%1s%4d    %8.3f%8.3f%8.3f%6.2f%6.2f"

    os.environ['GMX_SUPPRESS_DUMP'] = '1'  ## prevent gromacs filling the file system with step files

    aas = {'ALA':'A', 'ARG':'R', 'ASN':'N', 'ASP':'D', 'CYS':'C', 'GLN':'Q', 'GLU':'E', 
                'GLY':'G', 'HIS':'H', 'ILE':'I', 'LEU':'L', 'LYS':'K', 'MET':'M', 'PHE':'F', 
                'PRO':'P', 'SER':'S', 'THR':'T', 'TRP':'W', 'TYR':'Y', 'VAL':'V'}

    ### CG2AT folder locations

    start_dir       = os.getcwd()+'/'  ### initial working directory
    working_dir     = os.getcwd()+'/'+args.loc+'/'   ### working directory 
    final_dir       = os.getcwd()+'/'+args.loc+'/FINAL/'  ### final directory for run files
    input_directory = os.getcwd()+'/'+args.loc+'/INPUT/'  ### contains input run files
    merged_directory = os.getcwd()+'/'+args.loc+'/MERGED/'  ### contains run files
    scripts_dir     = os.path.dirname(os.path.realpath(__file__))+'/' ### contains script files
    database_dir    = str(Path(*Path(scripts_dir).parts[:-1]))+'/' ### contains database files
    box_vec = ''
    user_at_input = False
    ter_res = {}  ## contains the chain termini residue info e.g.  0:[ARG, PRO]
    system = {}  ## number of system components e.g. PROTEIN:2 POPE:10, POPG:20
    backbone_coords = {} ## CG coordinates of the backbone beads
    coord_atomistic = {} ## de_novo atomisitic information e.g. coord_atomistic[chain_count][residue_number][atom][info....]
    user_cys_bond = {} ## contains resid of disulphide bonds e.g. user_cys_bond[chain][[cys_resid,cys_resid], [cys_resid,cys_resid]])
    cg_residues = {} ## dictionary of CG beads eg cg_residues[residue type(POPE)][resid(1)][bead name(BB)][residue_name(PO4)/coordinates(coord)]
    seq_cg = {} ## CG sequence e.g. seq_cg[chain][sequence]
    seq_at = {} ## user AT sequence e.g. seq_at[chain][sequence]
    seq_cg_other = {} ## CG seq for linked non protein residues e.g. DNA
    tc = {} ## contains script timings
    atomistic_protein_input_raw = {} ## Raw user atomistic info coord_atomistic[chain_count][residue_number][atom][info....]
    atomistic_protein_input_aligned = {}
    chain_count = 0 ## number of user atomistic chains
    other_atomistic ={} ## 
    forcefield_available, fragments_available = '',''
    forcefield_location, forcefield = '','' ## forcefield info
    np_residues, p_residues, mod_residues, o_residues, sol_residues, ion_residues = [],[],[],[],[],[] ## fragment names
    np_directories, p_directories, mod_directories, o_directories, sol_directories, ion_directories = [],[],[],[],[],[]  ## fragment locations
    database_locations = [] ## grouped directories
    water, water_info = [],[] ## water information
    swap_dict ={} ## CG residue swap
    res_top, sorted_connect, hydrogen, heavy_bond, ions, at_mass = {},{},{},{},[],{} ### topology information
    group_chains = None
    np_blocks = {}
    skip_disul = {}  ## contains info on whether user supplied complete chain information   
    alt_res_name = {} ## alternative residue names
    hydration = {}
    get_forcefield = True
    cg_chain_group = {}



#!/usr/bin/env python3

import os, sys
import numpy as np
import math
from distutils.dir_util import copy_tree
import multiprocessing as mp
import distutils.spawn
from shutil import copyfile
import glob
import re
import copy
import ntpath
import g_var


def check_alternate_resname(resname):
    if resname in g_var.alt_res_name:
        return  g_var.alt_res_name[resname]
    elif resname not in g_var.res_top:
        sys.exit('The residue '+resname+' cannot be found in the topology or alternate resnames')
    else:
        return resname

def fetch_forcefield_water_info():
    if g_var.args.info or g_var.args.posre != None or g_var.args.compare != None:
        g_var.get_forcefield = False


def forcefield_selection(test=False):
    ##### forcefield selection
    if g_var.get_forcefield:
        if g_var.args.ff != None:
            if os.path.exists(g_var.args.ff):
                g_var.forcefield_location, g_var.forcefield = path_leaf(g_var.args.ff)
                folder_copy_and_check(g_var.args.ff, g_var.final_dir+g_var.forcefield)
                g_var.opt['ff'] = g_var.forcefield
                if not test:
                    print('\nYou have chosen to use your own force field: '+g_var.forcefield+' in '+g_var.forcefield_location)
            elif path_leaf(g_var.args.ff)[1]+'.ff' in g_var.forcefield_available:
                forcefield_number = g_var.forcefield_available.index(path_leaf(g_var.args.ff)[1]+'.ff')
            elif path_leaf(g_var.args.ff)[1] in g_var.forcefield_available:
                forcefield_number = g_var.forcefield_available.index(path_leaf(g_var.args.ff)[1])
            else:
                print('Cannot find force field: '+g_var.args.ff+'  please select one from below\n')   
        if 'forcefield_number' not in locals() and g_var.forcefield == '':
            forcefield_number = database_selection(g_var.forcefield_available, 'force fields', test)    
        if g_var.forcefield == '':
            if not test:
                print('\nYou have selected the force field: '+g_var.forcefield_available[forcefield_number].split('.')[0])
            folder_copy_and_check(g_var.database_dir+'/forcefields/'+g_var.forcefield_available[forcefield_number], g_var.final_dir+g_var.forcefield_available[forcefield_number])
            g_var.forcefield_location, g_var.forcefield=g_var.database_dir+'forcefields/', g_var.forcefield_available[forcefield_number]
            g_var.opt['ff'] = g_var.forcefield_available[forcefield_number]

def get_rtp_database(ff_files):
    rtp = [files for files in ff_files if files.endswith('.rtp')]
    exclusions_rtp = ['bondedtypes','atoms','bonds','exclusions', 'angles', 'dihedrals', 'impropers','cmap']
    rtp_database = []
    for rtp_file in rtp:
        # rtp_database[rtp_file[:-4]]=[]
        with open(g_var.final_dir+g_var.forcefield+'/'+rtp_file, 'r') as rtp_input:
            for line in rtp_input.readlines():
                if not (line.isspace() or len(line) == 0) and line[0] not in [';', '#']:
                    if line.strip().startswith('[') and strip_header(line) not in exclusions_rtp: 
                        rtp_database.append(strip_header(line))
    return rtp_database

def get_tdb_database(ff_files, rtp_database):
    tdb = [files for files in ff_files if files.endswith('.tdb')]
    exclusions_tdb = ['None', 'replace', 'add', 'delete', 'impropers']
    tdb_database={}
    for tdb_file in tdb:
        # if tdb_file[:-6] not in tdb_database:
        #     tdb_database[tdb_file[:-6]]={}
        with open(g_var.final_dir+g_var.forcefield+'/'+tdb_file, 'r') as tdb_input:
            if tdb_file[-5] not in tdb_database:
                tdb_database[tdb_file[-5]]=[]
            for line in tdb_input.readlines():
                if not (line.isspace() or len(line) == 0) and line[0] not in [';', '#']: 
                    if line.strip().startswith('[') and strip_header(line) not in exclusions_tdb :
                        tdb_database[tdb_file[-5]].append(strip_header(line))
    return tdb_database


def get_termini_selections():
    ff_files = os.listdir(g_var.final_dir+g_var.forcefield)
    rtp_database = get_rtp_database(ff_files)
    tdb_database = get_tdb_database(ff_files, rtp_database)
    g_var.termini_selections = {'n': {}, 'c':{}}
    for tdb, ter in tdb_database.items():
            if len(ter) > 0:
                for res in rtp_database :
                    g_var.termini_selections[tdb][res]=[]
                    skip = []
                    specific=[]
                    generic=[]
                    numeric=[]
                    for t in ter:
                        if t.endswith('-') :
                            if t.startswith(res) and t.count('-') > 1:
                                specific.append(t)
                            elif t.count('-') == 1:
                                generic.append(t)
                        elif '-' in t and len(t.split('-')) > 0 and not t.endswith('-'):
                            if t.startswith(res):
                                specific.append(t)
                                if t.split('-')[-1].endswith('+') or t.split('-')[-1].endswith('-'):
                                    skip.append(t.split('-')[-1][:-1])
                                else:
                                    skip.append(t.split('-')[-1])
                        else:
                            cont=True
                            for s in skip:
                                if not t.startswith(s):
                                    cont = True
                                else:
                                    cont=False 
                                    break
                            if cont:
                                if not t[0].isalpha():
                                    numeric.append(t) 
                                else:             
                                    generic.append(t)               
                    g_var.termini_selections[tdb][res]=specific+generic+numeric+['None']            
            
            
def fragment_selection(test=False):
    ##### fragment selection
    frag_location, fragment_number, fragments_available_other = [],[],[]
    if g_var.args.fg != None:
        for frag_val, frag_path in enumerate(g_var.args.fg):
            if os.path.exists(frag_path):
                frag_loc, fragments = path_leaf(frag_path)
                frag_location.append(os.path.abspath(frag_loc)+'/')
                fragment_number.append(frag_val)
                fragments_available_other.append(fragments)
    if len(fragment_number) == 0:
        fragment_number = fetch_frag_number(g_var.fragments_available, test)
        frag_location = [g_var.database_dir+'fragments/']*len(g_var.fragments_available)
        if g_var.args.fg is None:
            g_var.opt['fg'] = []
            for database in fragment_number:
                g_var.opt['fg'].append(g_var.fragments_available[database])
    else:
        g_var.fragments_available = fragments_available_other
    fetch_residues(frag_location, g_var.fragments_available, fragment_number, test)

def correct_number_cpus():
    if g_var.args.ncpus != None:
        if g_var.args.ncpus > mp.cpu_count():
            print('you have selected to use more CPU cores than are available: '+str(g_var.args.ncpus))
            print('defaulting to the maximum number of cores: '+str(mp.cpu_count()))
            g_var.args.ncpus = mp.cpu_count()
    else:
        if mp.cpu_count() >= 8:
            g_var.args.ncpus = 8
        else:
            g_var.args.ncpus = mp.cpu_count()
    g_var.opt['ncpus'] = g_var.args.ncpus

def check_input_flag():
    if g_var.get_forcefield and g_var.args.c is None:
        g_var.parser.print_help(sys.stderr)
        sys.exit('\nError: the following arguments are required: -c\n')

def path_leaf(path):
    head, tail = ntpath.split(path)
    if not tail:
        return path.replace(ntpath.basename(head)+'/', ''), ntpath.basename(head)
    else:
        return path.replace(tail, ''), tail

### finds gromacs installation
def find_gromacs():
    if g_var.args.gmx != None:
        g_var.args.gmx=distutils.spawn.find_executable(g_var.args.gmx)
    else:
        g_var.args.gmx=distutils.spawn.find_executable('gmx')
    if g_var.args.gmx is None or type(g_var.args.gmx) != str:
        if os.environ.get("GMXBIN") != None:
            for root, dirs, files in os.walk(os.environ.get("GMXBIN")):
                for file_name in files:
                    if file_name.startswith('gmx') and file_name.islower() and '.' not in file_name:
                        g_var.args.gmx=distutils.spawn.find_executable(file_name)
                        if type(g_var.args.gmx) == str and g_var.args.gmx != None :
                            break
                        else:
                            g_var.args.gmx=None
                break
        if g_var.args.gmx is None:
            sys.exit('Cannot find gromacs installation')
    g_var.opt['gmx'] = g_var.args.gmx

def trunc_coord(xyz):
    xyz_new = []
    for coord in xyz:
        if len(str(coord)) > 8:
            if '.' in str(coord):
                xyz_new.append(np.round(coord, 7-len(str(int(coord)))))
            else:
                # print(np.round(coord, 1), coord, 9-len(str(int(coord))))
                xyz_new.append(np.round(coord, 8-len(str(int(coord)))))
        else:
            xyz_new.append(coord)
    return xyz_new[0],xyz_new[1],xyz_new[2]

def calculate_distance(p1, p2):
    return np.sqrt(((p1[0]-p2[0])**2)+((p1[1]-p2[1])**2)+((p1[2]-p2[2])**2))

def file_copy_and_check(file_in,file_out):
    if not os.path.exists(file_out) and os.path.exists(file_in):
        copyfile(file_in, file_out)

def folder_copy_and_check(folder_in,folder_out):
    if not os.path.exists(folder_out):
        copy_tree(folder_in, folder_out)

def flags_used():
    print('\nAll variables supplied have been saved in : \n'+g_var.input_directory+'script_inputs.dat')
    os.chdir(g_var.input_directory)
    with open('script_inputs.dat', 'w') as scr_input:
        scr_input.write('\n'+g_var.opt['input']+'\n')
        for var in g_var.opt:
            if var != 'input':
                scr_input.write('{0:9}{1:15}\n'.format(var,str(g_var.opt[var])))

def is_hydrogen(atom):
    if str.isdigit(atom[0]) and atom[1] != 'H':
        return False
    elif not str.isdigit(atom[0]) and not atom.startswith('H'):
        return False
    else:
        return True

def fetch_chain_groups():
    if g_var.args.group != None:
        if g_var.args.group[0] not in ['all','chain']:
            g_var.group_chains = {}
            for group_val, group in enumerate(g_var.args.group):
                for chain in group.split(','):
                    g_var.group_chains[int(chain)]=group_val 
        else:
            g_var.group_chains =  g_var.args.group[0]   

def split_swap(swap):

    if len(re.split(':', swap)) > 2:
        res_range = re.split(':', swap)[2].split(',')
        res_id = []
        for resid_section in res_range:
            if 'ALL' in resid_section.upper():
                return 'ALL', 'ALL'
            else:
                if '-' in resid_section:
                    spt = resid_section.split('-')
                    for res in range(int(spt[0]), int(spt[1])+1):
                        res_id.append(res)
                else:
                    res_id.append(int(resid_section))
        return res_range, res_id
    else:
        return 'ALL', 'ALL'

def sort_swap_group():
    if g_var.args.swap != None:
        for swap in g_var.args.swap:
            res_s = re.split(':', swap)[0].split(',')
            if re.split(':', swap)[1].split(',') is not type(int):
                res_e = re.split(':', swap)[1].split(',')
            else:
                sys.exit('swap layout is not correct')
            
            if len(res_s) == len(res_e):
                res_range, res_id = split_swap(swap)
                if res_s[0] not in g_var.swap_dict:
                    g_var.swap_dict[res_s[0]]={}
                if len(res_s) == 1:
                    g_var.swap_dict[res_s[0]][res_s[0]+':'+res_e[0]]={'ALL':'ALL'}
                else:
                    g_var.swap_dict[res_s[0]][res_s[0]+':'+res_e[0]]={}
                    for bead in range(1,len(res_s)):
                        g_var.swap_dict[res_s[0]][res_s[0]+':'+res_e[0]][res_s[bead]]=res_e[bead]
                g_var.swap_dict[res_s[0]][res_s[0]+':'+res_e[0]]['resid']=res_id
                g_var.swap_dict[res_s[0]][res_s[0]+':'+res_e[0]]['range']=res_range
            else:
                sys.exit('The length of your swap groups do not match')
        
def print_swap_residues():
    if g_var.args.swap != None:
        to_print = '\nYou have chosen to swap the following residues\n\n'
        to_print += '{0:^10}{1:^5}{2:^11}{3:^11}{4:^11}{5:^11}\n'.format('residue', 'bead', '     ', 'residue', 'bead', 'range')
        to_print += '{0:^10}{1:^5}{2:^11}{3:^11}{4:^11}{5:^11}\n'.format('-------', '----', '     ', '-------', '----', '-----')
        for residue in g_var.swap_dict:
            for swap in g_var.swap_dict[residue]:
                bead_s, bead_e='', ''
                for bead in g_var.swap_dict[residue][swap]:
                    if bead not in ['resid', 'range']:
                        bead_s+=bead+' '
                        bead_e+=g_var.swap_dict[residue][swap][bead]+' '
                    elif bead == 'range':
                        if g_var.swap_dict[residue][swap]['range'] != 'ALL':
                            ran=''
                            for resid_section in g_var.swap_dict[residue][swap]['range']:
                                ran += resid_section+', '
                            ran = ran[:-2]
                        else:
                            ran = g_var.swap_dict[residue][swap]['range']
                to_print += '{0:^10}{1:^5}{2:^11}{3:^11}{4:^11}{5:^11}\n'.format(swap.split(':')[0], bead_s, ' --> ', swap.split(':')[1], bead_e, ran)
        return to_print
    else:
        return ''
    

def new_box_vec(box_vec, box):
    box_vec_split = box_vec.split()[1:]
    box_vec_values, box_shift = [], []
    for xyz_val, xyz in enumerate(box):
        if xyz != 0:
            box_shift.append((float(box_vec_split[xyz_val])/2) - (float(xyz)/2))
            box_vec_values.append(np.round(float(xyz), 3))
        else:
            box_shift.append(0)
            box_vec_values.append(float(box_vec_split[xyz_val]))
    box_vec = g_var.box_line%(float(box_vec_values[0]), float(box_vec_values[1]), float(box_vec_values[2]), float(box_vec_split[3]), float(box_vec_split[4]),float(box_vec_split[5]))
    return box_vec, np.array(box_shift)

def strip_header(line):
    line_new = line.replace('[','').split(']', 1)[0]
    if len(line_new.split())>1 or len(line_new.split())==0:
        sys.exit('There is a issue in one of the fragment headers: \n'+line)
    return line_new.strip()

def topology_header(line, topology, location):
    top = strip_header(line).upper()
    if top in topology:
        return top
    else:
        print('The topology header line is incorrect, therefore ignoring: \n'+location+'.top')
        return ''  

def sep_fragments_topology(residue, location):
    topology={}
    group = 1
    topology = copy.deepcopy(g_var.topology)
    if os.path.exists(location+'.top'):
        with open(location+'.top', 'r') as top_input:
            for line_nr, line in enumerate(top_input.readlines()):
                if not line.startswith('#') and len(line) > 0:
                    if line.startswith('['):
                        topology_function = topology_header(line, topology, location)
                    else:
                        line_sep = line.split()
                        if len(line_sep) > 0:
                            if topology_function == 'GROUPS':
                                topology, group = add_groups(topology, line_sep, group)
                            elif topology_function in ['N_TERMINAL', 'C_TERMINAL']:
                                topology[topology_function] = ''.join(line_sep)
                            elif topology_function == 'CHIRAL':
                                topology = add_chiral(topology, line_sep)
                            elif topology_function == 'CONNECT':
                                topology = add_connections(topology, line_sep)
                            elif topology_function == 'ALT_RES':
                                sort_alternate_residues(line_sep, residue)
                            elif topology_function == 'HYDRATION':
                                sort_hydration(line_sep, residue,)
    return topology

def add_groups(topology, line_sep, group):
    if not g_var.args.mod:
        for bead in line_sep:
            topology['GROUPS'][bead] = group
        group += 1
    topology['GROUPS']['group_max'] = group
    return topology, group

def add_chiral(topology, line_sep):
    if len(line_sep) == 5:
        topology['CHIRAL']['atoms']+=line_sep
        topology['CHIRAL'][line_sep[0]]={'m':line_sep[1], 'c1':line_sep[2], 'c2':line_sep[3], 'c3':line_sep[4]}
    else:
        print('The chiral group section is incorrect: \n'+location+'.top')
    return topology

def add_connections(topology, line_sep):
    if len(line_sep) == 4:
        topology['CONNECT']['atoms'][line_sep[1]] = int(line_sep[3])
        if line_sep[0] in topology['CONNECT']:
            topology['CONNECT'][line_sep[0]]['atom']+=[line_sep[1]]
            topology['CONNECT'][line_sep[0]]['Con_Bd']+=[line_sep[2]]
            topology['CONNECT'][line_sep[0]]['dir']+=[int(line_sep[3])]
        else:
            topology['CONNECT'][line_sep[0]]={'atom':[line_sep[1]], 'Con_Bd':[line_sep[2]], 'dir':[int(line_sep[3])]}
    else:
        print('The bead connection group section is incorrect: \n'+location+'.top')
    return topology

def sort_alternate_residues(line_sep, residue):
    for alt_res in line_sep:
        if alt_res not in g_var.alt_res_name:
            g_var.alt_res_name[alt_res] = residue
        else:
            sys.exit('The alternate residue name: \"'+alt_res+'\" already corresponds to: \"'+
                     g_var.alt_res_name[alt_res]+'\"')

def sort_hydration(line_sep, residue):
    if len(line_sep) == 1 and residue not in g_var.hydration:
        g_var.hydration[residue] = line_sep[0]
    else:
        print('There is a issue with the hydration section of: ', residue)

def get_fragment_topology(residue, location):
    topology = sep_fragments_topology(residue, location[:-4])
    g_var.res_top[residue] = {'ATOMS': [], 'C_TERMINAL':topology['C_TERMINAL'], 'N_TERMINAL':topology['N_TERMINAL'], \
                             'CHIRAL':topology['CHIRAL'], 'GROUPS':{}, 'CONNECT':topology['CONNECT']}
    with open(location, 'r') as pdb_input:
        group=topology['GROUPS']['group_max']
        atom_list=[]
        grouped_atoms={}
        for line_nr, line in enumerate(pdb_input.readlines()):
            if line.startswith('['):
                bead = strip_header(line)
                if bead in topology['GROUPS']:
                    group_temp = topology['GROUPS'][bead]
                else:
                    group_temp = group
                    group+=1
                if group_temp not in grouped_atoms:
                    grouped_atoms[group_temp]={bead:[]}
                    g_var.res_top[residue]['GROUPS'][bead]=group_temp
                else:
                    grouped_atoms[group_temp][bead]=[]
                    g_var.res_top[residue]['GROUPS'][bead]=group_temp
            if line.startswith('ATOM'):
                line_sep = pdbatom(line)
                grouped_atoms[group_temp][bead].append(line_sep['atom_number'])
            if 'bead' not in locals():
                sys.exit('error reading:\n'+location)
            ### return backbone info for each aminoacid residue
            if bead in g_var.res_top[residue]['CONNECT']:
                if line.startswith('ATOM'):
                    line_sep = pdbatom(line)
                    if not is_hydrogen(line_sep['atom_name']):
                        atom_list.append(line_sep['atom_name'])    ### list of backbone heavy atoms
                g_var.res_top[residue]['ATOMS']=atom_list
    return grouped_atoms   

def sort_connectivity(atom_dict, heavy_bond):
    cut_group = {}
    if len(atom_dict) > 1:
        for group in atom_dict:
            cut_group[group]={}
            group_atoms = [atom for frag in atom_dict[group] for atom in atom_dict[group][frag] if atom in heavy_bond]
            non_self_group = [x for x in atom_dict.keys() if x != group]
            for atom in group_atoms:
                for bond in heavy_bond[atom]:
                    for group_2 in non_self_group:
                        for frag in atom_dict[group_2]:                          
                            if bond in atom_dict[group_2][frag]:
                                cut_group[group][atom] = [frag]
    return cut_group

def fetch_fragment_multi():
#### fetches the Backbone heavy atoms and the connectivity with pre/proceeding residues 
    amino_acid_itp = fetch_amino_rtp_file_location(g_var.forcefield_location+g_var.forcefield) 
    g_var.at_mass = fetch_atom_masses(g_var.forcefield_location+g_var.forcefield)
    for residue_type in [g_var.p_directories, g_var.o_directories, g_var.mod_directories]:
        for directory in range(len(residue_type)):
            for residue in residue_type[directory][1:]:    
                if residue not in g_var.res_top:
                    location = fragment_location(residue)
                    grouped_atoms = get_fragment_topology(residue, location)
                    g_var.hydrogen[residue], g_var.heavy_bond[residue], residue_list, at_mass, amide_h = fetch_bond_info(residue, amino_acid_itp, g_var.at_mass, location)
                    
                    g_var.sorted_connect[residue]  = sort_connectivity(grouped_atoms, g_var.heavy_bond[residue])
                    g_var.res_top[residue]['RESIDUE'] = residue_list
                    g_var.res_top[residue]['atom_masses'] = at_mass
                    if residue in g_var.p_residues:
                        g_var.res_top[residue]['amide_h'] = amide_h
                    if residue in g_var.alt_res_name:
                        g_var.res_top[g_var.alt_res_name[residue]]=g_var.res_top[residue]

def fetch_fragment_single():
    for frag_val, frag_dir in enumerate([g_var.sol_directories, g_var.ion_directories, g_var.np_directories]):
        for directory in range(len(frag_dir)):
            for residue in frag_dir[directory][1:]:    
                if residue not in g_var.res_top:
                    location = fragment_location(residue)
                    grouped_atoms = get_fragment_topology(residue, location)
                    if frag_val <=1: 
                        at_mass = fetch_atoms_water_ion(frag_dir[directory][0]+residue+'/', g_var.at_mass)
                        g_var.hydrogen[residue], g_var.heavy_bond[residue], g_var.sorted_connect[residue] = {},{},{}
                        residue_list = [residue]
                    else:
                        g_var.hydrogen[residue], g_var.heavy_bond[residue], residue_list, at_mass, amide_h  = fetch_bond_info(residue, [location[:-4]+'.itp'], g_var.at_mass, location)
                        g_var.sorted_connect[residue]  = sort_connectivity(grouped_atoms, g_var.heavy_bond[residue])
                     
                    g_var.res_top[residue]['RESIDUE'] = residue_list
                    g_var.res_top[residue]['atom_masses'] = at_mass

def atom_bond_check(line_sep):
    if line_sep[1] == 'atoms':
        return True, False
    elif line_sep[1] == 'bonds':
        return False, True
    else:
        return False, False

def fetch_amino_rtp_file_location(forcefield_loc):
    rtp=[]
    for file in os.listdir(forcefield_loc):
        if file.endswith('.rtp'):
            rtp.append(forcefield_loc+'/'+file)
    return rtp

def fetch_atom_masses(forcefield_loc):
    at_mass = {}
    if os.path.exists(forcefield_loc+'/atomtypes.atp'):
        with open(forcefield_loc+'/atomtypes.atp', 'r') as itp_input:
            for line in itp_input.readlines():
                if not (line.isspace() or len(line) == 0) and line[0] not in [';', '#']:
                    line_sep = line.split()       
                    at_mass[line_sep[0]]=line_sep[1] 
    else:
        sys.exit('cannot find atomtypes.dat in the force field: '+forcefield_loc)
    return at_mass

def fetch_atoms_water_ion(forcefield_loc, at_mass_p):
    at_mass = {}
    for file in os.listdir(forcefield_loc):
        if file.endswith('itp'):
            with open(forcefield_loc+file, 'r') as itp_input:
                strip_atoms = False
                for line in itp_input.readlines():
                    line_sep = line.split()
                    if not (line.isspace() or len(line) == 0) and line[0] not in [';', '#']:
                        if line.strip().startswith('[') and strip_header(line) == 'atoms':
                            strip_atoms = True
                        elif line.strip().startswith('['):
                            strip_atoms = False
                        elif strip_atoms:
                            at_mass[line_sep[4]] = float(at_mass_p[line_sep[1]])
    return at_mass


def check_res_name(residue, residue_itp):
    if residue_itp in g_var.alt_res_name or residue_itp == residue:
        if residue_itp == residue:
            return True
        elif g_var.alt_res_name[residue_itp] == residue:
            return True
        else:
            return False
    else:
        return False


def fetch_bond_info(residue, rtp, at_mass, location):
    bond_dict=[]
    heavy_dict, H_dict=[],[]
    residue_present, mol_type = False,False
    atom_conversion = {}
    residue_list=[]
    res_at_mass = {}
    for rtp_file in rtp:
        with open(rtp_file, 'r') as itp_input:
            for line in itp_input.readlines():
                if not (line.isspace() or len(line) == 0) and line[0] not in [';', '#']:
                    line_sep = line.split()
                    if line.strip().startswith('[') and not residue_present: 
                        residue_present = check_res_name(residue, strip_header(line))  
                        if strip_header(line) == 'moleculetype':
                            mol_type = True
                    elif mol_type:
                        residue_present = check_res_name(residue, line.split()[0])  
                        mol_type = False
                    elif residue_present or residue not in g_var.p_residues+g_var.o_residues:
                        if line.strip().startswith('[') :
                            atoms, bonds = atom_bond_check(line_sep)
                        elif 'atoms' in locals():
                            if atoms and residue in g_var.p_residues + g_var.o_residues:
                                residue_list, atom_conversion, H_dict, res_at_mass, heavy_dict = fetch_bond_info_atoms_linked(residue, line_sep, residue_list, atom_conversion, H_dict, res_at_mass, heavy_dict)
                            elif atoms:
                                residue_list, H_dict, res_at_mass, heavy_dict = fetch_bond_info_atoms_NP(residue, line_sep, residue_list, H_dict, res_at_mass, heavy_dict)
                            elif bonds:
                                try:
                                    bond_dict.append([int(line_sep[0]),int(line_sep[1])])
                                except:
                                    bond_dict.append([line_sep[0],line_sep[1]])
                            elif not atoms and not bonds and residue in g_var.p_residues+g_var.o_residues:
                                break
            # if 'atoms' not in locals():
            #     print('Issue finding information for residue: ',residue)
            #     sys.exit('There is a issue with: \n'+rtp_file)
        if len(heavy_dict) > 0:
            break
    if 'atoms' not in locals():
        print('Issue finding information for residue: ',residue)
        sys.exit('There is a issue with: \n'+rtp_file)
    if not residue_present:
        sys.exit('cannot find topology information for: '+residue)
    bond_dict=np.array(bond_dict)
    hydrogen = {}
    heavy_bond = {}
    if residue not in g_var.mod_residues:
        atom_conversion = get_atomistic(location)
    for bond in bond_dict:
        hydrogen, amide_h = add_to_topology_list(bond[0], bond[1], hydrogen, heavy_dict, H_dict, atom_conversion, residue, g_var.p_residues+g_var.o_residues)
        if residue in g_var.p_residues and amide_h != None:
            amide_hydrogen = amide_h
        heavy_bond, amide_h= add_to_topology_list(bond[0], bond[1], heavy_bond, heavy_dict, heavy_dict, atom_conversion, residue, g_var.p_residues+g_var.o_residues)
    if 'amide_hydrogen' in locals():
        return hydrogen, heavy_bond, residue_list, res_at_mass, amide_hydrogen
    else:
        return hydrogen, heavy_bond, residue_list, res_at_mass, None

def fetch_bond_info_atoms_linked(residue, line_sep, residue_list, atom_conversion, H_dict, res_at_mass, heavy_dict):
    if residue not in residue_list:
        residue_list.append(residue)
    atom_conversion[line_sep[0]]=int(line_sep[3])+1
    if is_hydrogen(line_sep[0]):
        H_dict.append(line_sep[0])
    else:
        res_at_mass[line_sep[0]] = float(g_var.at_mass[line_sep[1]])
        heavy_dict.append(line_sep[0])
    return residue_list, atom_conversion, H_dict, res_at_mass, heavy_dict

def fetch_bond_info_atoms_NP(residue, line_sep, residue_list, H_dict, res_at_mass, heavy_dict):
    if line_sep[3] not in residue_list:
        residue_list.append(line_sep[3])
    if is_hydrogen(line_sep[4]):
        H_dict.append(int(line_sep[0]))
    else:
        res_at_mass[line_sep[4]] = float(line_sep[7])
        heavy_dict.append(int(line_sep[0]))
    return residue_list, H_dict, res_at_mass, heavy_dict


def get_atomistic(frag_location):
#### read in atomistic fragments into dictionary    
    residue = {} 
    with open(frag_location, 'r') as pdb_input:
        for line_nr, line in enumerate(pdb_input.readlines()):
            if line.startswith('ATOM'):
                line_sep = pdbatom(line) ## splits up pdb line
                residue[line_sep['atom_name']] = line_sep['atom_number']
    return residue


def add_to_topology_list(bond_1, bond_2, top_list, dict1, dict2, conversion, residue, linked_residues):
    amide_hydrogen = None
    for bond in [[bond_1, bond_2], [bond_2, bond_1]]:
        if bond[0] in dict1 and bond[1] in dict2:
            if residue in linked_residues:
                if bond[0] == 'N' and is_hydrogen(bond[1]):
                    amide_hydrogen = bond[1]
                if bond[0] in conversion and bond[1] in conversion:
                    bond[0], bond[1] = conversion[bond[0]],conversion[bond[1]] 
            if bond[0] not in top_list:
                top_list[bond[0]]=[]
            top_list[bond[0]].append(bond[1])
    return top_list, amide_hydrogen

def fragment_location(residue):  
#### runs through dirctories looking for the atomistic fragments returns the correct location
    for res_type in [g_var.np_directories, g_var.p_directories, g_var.mod_directories, g_var.o_directories, g_var.sol_directories, g_var.ion_directories]:
        for directory in range(len(res_type)):
            if os.path.exists(res_type[directory][0]+residue+'/'+swap_to_solvent(residue)+'.pdb'):
                return res_type[directory][0]+residue+'/'+swap_to_solvent(residue)+'.pdb'            
    sys.exit('Cannot find fragment: '+residue+'/'+swap_to_solvent(residue)+'.pdb')


def read_database_directories():
#### Read in forcefields provided
    available_provided_database=[]
    for directory_type in ['forcefields', 'fragments']:
        if os.path.exists(g_var.database_dir+directory_type):
            for root, dirs, files in os.walk(g_var.database_dir+directory_type):
                available_provided = [x for x in sorted(dirs) if not x.startswith('_')]
                break
        else:
            sys.exit('no '+directory_type+' found')
            available_provided=[]
        available_provided_database.append(available_provided)
    g_var.forcefield_available, g_var.fragments_available = available_provided_database[0], available_provided_database[1]


def database_selection(provided, selection_type, test=False):
#### print out selection of forcefields
    if not test:
        print('\n\n{0:^79}\n'.format('Provided '+selection_type))
        print('{0:^90}'.format('{0:^20}{1:^41}'.format('Selection',selection_type)))
        print('{0:^90}'.format('{0:^20}{1:^41}'.format('---------','-'*len(selection_type))))
        for force_num_prov, line in enumerate(provided):
            print('{0:^90}'.format('{0:^20}{1:^41}'.format(force_num_prov,line.split('.')[0])))
    return ask_database(provided,  selection_type)

def ask_database(provided, selection_type, test=False):
#### ask which database to use
    attempt=0
    while True:
        try:
            if len(provided)==1:
                print('\nOnly 1 '+selection_type[:-1]+' database is currently available, therefore you have no choice but to accept the following choice.')
                number = 0
                return number
        #### if asking about fragments accept a list of libraries 
            if selection_type=='fragments': 
                number = np.array(input('\nplease select fragment libraries (in order of importance: eg. "1 0" then ENTER): ').split())
                number=number.astype(int)
                if len(number[np.where(number >= len(provided))]) == 0:
                    return number
                elif test:
                    return True
        #### if forcefield only accept one selection
            else:
                number = int(input('\nplease select a force field: '))
                if number < len(provided):
                    return number
                elif test:
                    return True
        except KeyboardInterrupt:
            sys.exit('\nInterrupted')
        except BaseException:
            if test:
                return True
            print("Oops!  That was a invalid choice")
            attempt+=1
            if attempt > 3:
                sys.exit('Too many invalid choices')

def fetch_frag_number(fragments_available, test=False):
    fragment_number = []
    if g_var.args.fg != None and len(g_var.args.fg) > 0 :
        for frag in g_var.args.fg:
            if frag in fragments_available:
                fragment_number.append(fragments_available.index(frag))
            elif not g_var.args.info:
                print('Cannot find fragment library: '+frag+' please select library from below\n')
                fragment_number += database_selection(fragments_available, 'fragments', test).tolist()
    else:
        fragment_number = database_selection(fragments_available, 'fragments', test)
    if len(fragment_number) > 0:
        return fragment_number 
    else:
        if g_var.args.info:
            return []
        sys.exit('no fragment databases selected')

def add_to_list(root, dirs, list_to_add, residues):
    list_to_add.append([])
    list_to_add[-1].append(root)
    list_to_add[-1] += dirs
    list1 = [x for x in list_to_add[-1] if not x.startswith('_')]
    list1.sort()
    list_to_add[-1] = list1
    residues += list_to_add[-1][1:]
    residues.sort()    

def fetch_residues(frag_dir, fragments_available_prov, fragment_number, test=False):
#### list of directories and water types  [[root, folders...],[root, folders...]]
#### run through selected fragments
    for database in fragment_number:
        if not g_var.args.info and not test:
            if g_var.database_dir in frag_dir[database]:
                print('\nYou have selected the fragment library: '+fragments_available_prov[database])
            else:
                print('\nYou have chosen to use your own fragment library: '+fragments_available_prov[database]+' in '+frag_dir[database])
    #### separate selection between provided and user
        location = frag_dir[database]+ fragments_available_prov[database]
    #### runs through protein and non protein
        for directory_type in ['/non_protein/', '/protein/', '/other/', '/protein_modified/', '/solvent/', '/ions/']:
            if os.path.exists(location+directory_type):
                for root, dirs, files in os.walk(location+directory_type):
                    #### adds non protein residues locations to np_directories
                    if directory_type =='/non_protein/':
                        add_to_list(root, dirs, g_var.np_directories, g_var.np_residues)
                    #### adds other mutli residue locations to o_directories
                    elif directory_type =='/other/':
                        add_to_list(root, dirs, g_var.o_directories, g_var.o_residues)
                    #### adds protein residues locations to p_directories
                    elif directory_type =='/protein/':
                        add_to_list(root, dirs, g_var.p_directories, g_var.p_residues)
                    elif directory_type =='/protein_modified/':
                        add_to_list(root, dirs, g_var.mod_directories, g_var.mod_residues)
                        g_var.p_residues+=g_var.mod_residues
                    elif directory_type =='/ions/':
                        add_to_list(root, dirs, g_var.ion_directories, g_var.ion_residues)
                    elif directory_type =='/solvent/':
                        add_to_list(root, dirs, g_var.sol_directories, g_var.sol_residues)
                    break
            else:
                print('Cannot find fragments for: ', directory_type[1:-1] )
    

def print_water_selection(water):
    to_print =''
    if g_var.args.w != None:
        to_print +='\nThe water type '+g_var.args.w+' doesn\'t exist\n'
    to_print +='\nPlease select a water molecule from below:\n\n'
    to_print +='{0:^20}{1:^30}\n'.format('Selection','water_molecule')
    to_print +='{0:^20}{1:^30}\n'.format('---------','----------')
    for selection, water_model in enumerate(water):
        to_print +='{0:^20}{1:^30}\n'.format(selection,water_model)
    return to_print

def ask_for_water_model(water):
    attempt=0
    while True:
        try:
            number = int(input('\nplease select a water model: '))
            if number < len(water):
                return water[number]
        except KeyboardInterrupt:
            sys.exit('\nInterrupted')
        except BaseException:
            print("Oops!  That was a invalid choice")
            attempt+=1
            if attempt > 3:
                sys.exit('Too many invalid choices')

def check_water_molecules(test=False):
    water_info=[]        
    water=np.array([])
    for directory in g_var.sol_directories:
        water_info.append([directory[0]])
        for fragment in directory[1:]:
            for filename in os.listdir(directory[0]+fragment+'/'):
                if filename.endswith('.pdb') and not filename.startswith('_'):
                    water = np.append(water, filename[:-4])
                    water_info[-1].append(filename[:-4])
        water_info[-1] = np.sort(np.unique(np.array(water_info[-1])))
    g_var.water_info = water_info  
    g_var.water = np.sort(np.unique(water))
    if g_var.get_forcefield:
        if g_var.args.w != None:
            g_var.args.w=g_var.args.w.upper()
        if len(water) == 0:
            print('WARNING cannot find any solvent fragments')
        else:
            if g_var.args.w in water:
                print('\nYou have selected the water model: '+g_var.args.w)
            else:
                print(print_water_selection(g_var.water))
                g_var.args.w = ask_for_water_model(g_var.water)
            g_var.opt['w'] = g_var.args.w 

def swap_to_solvent(residue_type):
    if residue_type in g_var.sol_residues:
        return g_var.args.w
    else:
        return residue_type

############################################################################################## fragment rotation #################################################################################

def AnglesToRotMat(theta) :
#### rotation matrices for the rotation of fragments. theta is [x,y,z] in radians     
    R_x = np.array([[1,         0,                  0                   ],
                    [0,         math.cos(theta[0]), -math.sin(theta[0]) ],
                    [0,         math.sin(theta[0]), math.cos(theta[0])  ]
                    ])
         
    R_y = np.array([[math.cos(theta[1]),    0,      math.sin(theta[1])  ],
                    [0,                     1,      0                   ],
                    [-math.sin(theta[1]),   0,      math.cos(theta[1])  ]
                    ])
                 
    R_z = np.array([[math.cos(theta[2]),    -math.sin(theta[2]),    0],
                    [math.sin(theta[2]),    math.cos(theta[2]),     0],
                    [0,                     0,                      1]
                    ])
                                        
    R = np.dot(R_z, np.dot( R_y, R_x ))
    return R                          

def angle_clockwise(A, B):
#### find angle between vectors
    AB = np.linalg.norm(A)*np.linalg.norm(B)
    A_dot_B = A.dot(B)
    angle = np.degrees(np.arccos(A_dot_B/AB))
#### determinant of A, B
    determinant = np.linalg.det([A,B])

    if determinant < 0: 
        return angle
    else: 
        return 360-angle

############################################################################################## fragment rotation done #################################################################################

def pdbatom(line):
### get information from pdb file
### atom number, atom name, residue name,chain, resid,  x, y, z
    try:
        return dict([('atom_number',int(line[7:11].replace(" ", ""))),('atom_name',str(line[12:16]).replace(" ", "")),('residue_name',str(line[16:21]).replace(" ", "")),\
            ('chain',line[21]),('residue_id',int(line[22:26])), ('x',float(line[30:38])),('y',float(line[38:46])),('z',float(line[46:54]))])
    except BaseException:
        sys.exit('\npdb line is wrong:\t'+line) 

def create_pdb(file_name):
    pdb_output = open(file_name, 'w')
    pdb_output.write('TITLE     GENERATED BY CG2AT\nREMARK    Please don\'t explode\nREMARK    Good luck\n\
'+g_var.box_vec+'MODEL        1\n')
    return pdb_output

def mkdir_directory(directory):
#### checks if folder exists, if not makes folder
    if not os.path.exists(directory):
        os.mkdir(directory)

def clean(test=False):
#### cleans temp files from residue_types
    if not test:
        print()
    for residue_type in g_var.cg_residues:

        if not test:
            print('Cleaning temp files from : '+residue_type)
        os.chdir(g_var.working_dir+residue_type)
        file_list = glob.glob('*temp*', recursive=True)
        file_list += glob.glob(residue_type+'*pdb', recursive=True)
        for file in file_list:
            if not file.endswith('.tpr') and not file.endswith('_merged.pdb') and 'gmx' not in file:
                os.remove(file)
        os.chdir(g_var.working_dir+residue_type+'/MIN')
        file_list = glob.glob('*temp*', recursive=True)
        file_list += glob.glob('*trr', recursive=True)
        for file in file_list:
            if not file.endswith('.tpr'):
                os.remove(file) 

def fix_time(t1, t2):
    minutes, seconds= divmod(t1-t2, 60)
    if minutes > 60:
        hours, minutes = divmod(minutes, 60)
    else:
        hours = 0
    return '{0:^3}{1:^6}{2:^3}{3:^4}{4:^3}{5:^4}'.format(int(np.round(hours)),'hours',int(np.round(minutes)),'min',int(np.round(seconds,0)),'sec') 

def print_script_timings():
    to_print=[]
    to_print.append('\n{:-<100}'.format(''))
    to_print.append('\n{0:^47}{1:^22}'.format('Job','Time'))
    to_print.append('{0:^47}{1:^22}'.format('---','----'))
    to_print.append('\n{0:47}{1}'.format('Initialisation: ', fix_time(g_var.tc['i_t_e'],g_var.tc['i_t'])))
    to_print.append('{0:47}{1}'.format('Read in CG system: ', fix_time(g_var.tc['r_i_t'],g_var.tc['i_t_e']))) 
    if 'PROTEIN' in g_var.system:
        to_print.append('{0:47}{1}'.format('Build protein systems: ',fix_time(g_var.tc['f_p_t'],g_var.tc['r_i_t'])))
    if 'OTHER' in g_var.system:
        to_print.append('{0:47}{1}'.format('Build other systems: ',fix_time(g_var.tc['f_o_t'],g_var.tc['f_p_t'])))        
    to_print.append('{0:47}{1}'.format('Build non protein system: ',fix_time(g_var.tc['n_p_t'],g_var.tc['f_o_t'])))
    to_print.append('{0:47}{1}'.format('Merge and minimise de novo: ',fix_time(g_var.tc['m_t'],g_var.tc['n_p_t'])))
    to_print.append('{0:47}{1}'.format('NVT on de novo: ',fix_time(g_var.tc['eq_t'],g_var.tc['m_t'])))
    if g_var.args.o in ['all', 'align'] and g_var.user_at_input:
        to_print.append('{0:47}{1}'.format('Creating aligned system: ',fix_time(g_var.tc['a_e'],g_var.tc['a_s'])))
    to_print.append('{:-<69}'.format(''))
    to_print.append('{0:47}{1}'.format('Total run time: ',fix_time(g_var.tc['f_t'],g_var.tc['i_t'])))
    with open(g_var.final_dir+'script_timings.dat', 'w') as time_out:  
        for line in to_print:
            time_out.write(line+'\n')
            if g_var.args.v >= 1:
                print(line)
        print('\nAll script timings have been saved in: \n'+g_var.final_dir+'script_timings.dat\n')

def cg2at_header():
    print('\n{0:^90}\n'.format('CG2AT2 is a fragment based conversion of coarse-grained to atomistic.'))
    print('{0:^90}\n'.format('CG2AT2 version: '+str(g_var.version)+' lite (for CCD2MD)'))
    print('{0:^90}\n'.format('Last updated : '+str(g_var.script_update)))
    print('{0:^90}'.format('CG2AT2 is written by Owen Vickery'))
    print('{0:^90}'.format('Project leader Phillip Stansfeld'))
    print('\n{0:^90}\n{1:^90}'.format('Contact email address:','cg2at2@gmail.com'))
    print('\n{0:^90}\n{1:^90}\n{2:^90}\n{3:-<90}'.format('Address:','School of Life Sciences, University of Warwick,','Gibbet Hill Road, Coventry, CV4 7AL, UK', ''))
    print('{0:^90}'.format('Please email me any new residues for the database!'))
    print('\n{0:^90}\n{1:^90}'.format('If you are using this script please acknowledge me (Dr Owen Vickery)','and cite the following:'))    
    print('\n{0:^90}\n{1:^90}\n{2:^90}\n{3:^90}\n{4:^90}\n{5:^90}'.format('CG2AT2: an Enhanced Fragment-Based Approach for ','Serial Multi-scale Molecular Dynamics Simulations',\
        'Owen N. Vickery and Phillip J. Stansfeld','Journal of Chemical Theory and Computation','2021 17 (10), 6472-6482','DOI: 10.1021/acs.jctc.1c00295'))
    print('\n{0:-<90}\n{1:^90}'.format('', 'File locations'))
    print('\n{0:^90}'.format('Executable: '+g_var.opt['input'].split()[0]))
    print('{0:^90}'.format('Database locations: '+g_var.database_dir))
    print('{0:^90}\n\n{1:-<90}'.format('Script locations: '+g_var.scripts_dir, ''))

def database_information():
    
    to_print = '\n{0:^90}\n{1:-<90}\n\n'.format('The available force fields within your database are (flag -ff):', '')
    for forcefields in g_var.forcefield_available:
        to_print += '{0:^90}\n'.format(forcefields)
    to_print += '\n\n{0:^90}\n{1:-<90}\n\n'.format('The available fragment libraries within your database are (flag -fg):', '')
    for fragments in g_var.fragments_available:
        to_print += '{0:^90}\n'.format(fragments)   
    if g_var.args.fg != None :
        to_print = fragments_in_use(to_print)
    sys.exit(to_print+'\n\"If all else fails, immortality can always be assured by spectacular error.\" (John Kenneth Galbraith)\n')

def fragments_in_use(to_print=''):
    protein_directories=[]
    if np.any(np.array([g_var.np_directories, protein_directories, g_var.mod_directories, g_var.o_directories, g_var.water_info, g_var.ion_directories], dtype=object)):
        for database_val, database in enumerate(sorted(g_var.args.fg)):
            to_print += '\n\n{0:^90}\n{1:-<90}\n\n'.format('The following residues are available in the database: '+database,'')
            res_type_name = ['Non protein residues', 'Protein residues', 'Modified protein residues', 'Other linked residues', 'Solvent residues', 'Solvent models', 'Ions']
            for res_val, residue in enumerate([g_var.np_directories, g_var.p_directories, g_var.mod_directories, g_var.o_directories, g_var.sol_directories, g_var.water_info, g_var.ion_directories]):
                try:
                    res_type = sorted(residue[database_val][1:])
                    to_print += '\n{0:^90}\n{1:^90}\n'.format(res_type_name[res_val], '-'*len(res_type_name[res_val]))
                    if len(', '.join(map(str, res_type))) <= 80:
                        to_print += '{0:^90}\n'.format(', '.join(map(str, res_type)))
                    else:
                        start, end = 0, 1                       
                        while end < len(res_type):
                            line = ', '.join(map(str, res_type[start:end]))
                            while len(line) <= 80:
                                if end < len(res_type):
                                    end+=1
                                    line = ', '.join(map(str, res_type[start:end]))
                                    if len(line) > 80:
                                        end-=1
                                        line = ', '.join(map(str, res_type[start:end]))
                                        break
                                else:
                                    break
                            to_print += '{0:^90}\n'.format(line)
                            start = end
                except:
                    pass
        to_print += '\n{0:-<90}\n\n'.format('')
    return to_print

def write_system_components():
    to_write = '\n{:-<100}\n'.format('')
    to_write += '{0:^100}\n'.format('Script has completed, time for a beer')
    to_write += '\n{0:^10}{1:^25}\n'.format('molecules','number')
    to_write += '{0:^10}{1:^25}\n'.format('---------','------')
    for section in g_var.system:
        to_write += '{0:^10}{1:^25}\n'.format(section, g_var.system[section])
    return to_write

def print_sequnce_info(sys_type):
    sequence_info = [g_var.seq_cg[sys_type], g_var.seq_at[sys_type]] if g_var.user_at_input and sys_type == 'PROTEIN' else [g_var.seq_cg[sys_type]]
    to_print = ''
    for rep_val, rep in enumerate(sequence_info):
        if rep_val == 0 and len(rep) != 0:
            to_print += 'Summary of coarse-grained '+sys_type+' chains\n'
        elif rep_val > 0 and len(rep) != 0:
            to_print += '\nSummary of atomistic '+sys_type+' chains\n'
        else:
            break
        to_print += '\n{0:^15}{1:^12}\n'.format('chain number', 'length of chain') #   \nchain number\tDelta A\t\tno in pdb\tlength of chain')
        to_print += '\n{0:^15}{1:^12}\n'.format('------------', '---------------')
        for chain in rep:
            to_print += '{0:^15}{1:^12}\n'.format(chain, len(rep[chain]))
        to_print += '\nSequences:\n'
        counter=0
        for index in rep:
            rep, to_print, counter = print_sequnce_info_header(rep_val, rep, to_print, counter, index)
            to_print += '{0:9}{1:10}{2:10}{3:10}{4:10}{5:10}{6:10}{7:10}{8:10}{9:10}\n'.format('1','10','20','30','40','50','60','70','80','90')
            to_print = print_to_100_char(rep[index], to_print)
    return to_print

def print_to_100_char(list_to_print, to_print):
    if len(''.join(map(str, list_to_print))) <= 100:
        to_print += '{0:100}\n'.format(''.join(map(str, list_to_print)))
    else:
        start, end = 0, 1                       
        while end < len(list_to_print):
            line = ''.join(map(str, list_to_print[start:end]))
            while len(line) <= 100:
                if end < len(list_to_print):
                    end+=1
                    line = ''.join(map(str, list_to_print[start:end]))
                    if len(line) > 100:
                        end-=1
                        line = ''.join(map(str, list_to_print[start:end]))
                        break
                else:
                    break
            to_print += '{0:100}\n'.format(line)
            start = end
    return to_print

def print_sequnce_info_header(rep_val, rep, to_print, counter, index):
    if rep_val == 0:
        to_print += '\nCG chain: '+str(index)+'\n'
    else:
        to_print += '\nAT chain: '+str(index)+' -> Group '+str(g_var.group_chains[index])+' -> CG chain '+str(g_var.cg_chain_group[index])+'\n'
        if len(g_var.atomistic_protein_input_aligned[g_var.cg_chain_group[index]].keys()) > 1:
            for seq in list(g_var.atomistic_protein_input_aligned[g_var.cg_chain_group[index]].keys()):
                chain_max = int(seq.split(':')[1])
            chain_max = np.max(chain_max)
            seq_range = list(g_var.atomistic_protein_input_aligned[g_var.cg_chain_group[index]].keys())[counter].split(':')
            rep[index] = ['-']*int(seq_range[0])+rep[index]+['-']*(chain_max-int(seq_range[1]))
            counter += 1
        else:
            seq_range = list(g_var.atomistic_protein_input_aligned[g_var.cg_chain_group[index]].keys())[counter].split(':')
            if int(seq_range[0]) > 0:
                rep[index] = ['-']*int(seq_range[0])+rep[index]
            counter = 0
    return rep, to_print, counter



#!/usr/bin/env python3

import os, sys
import numpy as np
import itertools
import math
from scipy.spatial import cKDTree
import gen, g_var, at_mod_p, read_in, gro


### sanity checking

def sanity_check_fragments(res, cg, sin_bead):
#### fetches bead and atom info from fragment database
    location = gen.fragment_location(res)
    residue, fragment_mass = get_atomistic(location, res)
    atom_list = []
    bead_list = []
    for group in residue.values():
        for bead in group:
            bead_list.append(bead)
            if not sin_bead:
                for atom in group[bead]:
                    atom_list.append(atom)
            else:
                if bead == sin_bead:
                    for atom in group[bead]:
                        atom_list.append(atom)
    return sorted(bead_list), atom_list

def sanity_check_atoms(atom_list, res):
#### checks atom order
    for at_num in range(1, len(atom_list)+1):
        if at_num not in atom_list:
            sys.exit('atom number '+str(at_num)+' is missing from fragment library: '+res+'\n')

def sanity_check_beads(bead_list, cg, res):
#### checks if bead is in fragment library
    bead_list_new=[]
    for bead in cg:
        if bead not in bead_list:
            new_bead = bead[1:]+bead[0]
            if new_bead in bead_list and new_bead not in cg:
                pass
            else:
                sys.exit('The bead '+bead+' is missing from the fragment library: '+res+'\n')   
        bead_list_new.append(bead)
    return bead_list_new

def sanity_check_protein_other(res_type, test=False):
    bead_list, atom_list = {},{}
    for residue in g_var.cg_residues[res_type]:
        for bead in g_var.cg_residues[res_type][residue]:
            resname = g_var.cg_residues[res_type][residue][bead]['residue_name']
            break
        if resname not in bead_list or resname not in atom_list:
            bead_list[resname], atom_list[resname] = sanity_check_fragments(resname, g_var.cg_residues[res_type][residue], False)
            sanity_check_atoms(atom_list[resname], resname)
        bead_list_cg = sanity_check_beads(bead_list[resname], g_var.cg_residues[res_type][residue], resname)  
        if bead_list[resname] != sorted(bead_list_cg):
            if len(bead_list[resname]) == len(bead_list_cg):
                fix_atom_wrap(bead_list[resname], bead_list_cg, res_type, residue)
            else:
                if not test:
                    print('There is a issue with residue: '+resname+' '+str(residue+1)+'. If expected ignore this message.' )
                if res_type != 'OTHER':
                    sys.exit('number of atomistic fragments: '+str(len(bead_list[resname]))+' does not equal number of CG beads: '+str(len(bead_list_cg)))

def sanity_check_solvent(res_type):
    bead_list, atom_list = {},{}
    for residue in g_var.cg_residues[res_type]:
        for bead in g_var.cg_residues[res_type][residue]:
            sin_bead=bead
            break
        if sin_bead not in bead_list or sin_bead not in atom_list:
            bead_list[sin_bead], atom_list[sin_bead] = sanity_check_fragments(res_type, g_var.cg_residues[res_type][residue], sin_bead)
            sanity_check_atoms(atom_list[sin_bead], res_type)
        sanity_check_beads(bead_list[sin_bead], g_var.cg_residues[res_type][residue], res_type)

def sanity_check_non_protein(res_type):
    bead_list, atom_list = {},{}
    if res_type not in bead_list or res_type not in atom_list:
        bead_list[res_type], atom_list[res_type] = sanity_check_fragments(res_type, g_var.cg_residues[res_type], False)
    sanity_check_atoms(atom_list[res_type], res_type)
    for residue in g_var.cg_residues[res_type]:
        bead_list_cg = sanity_check_beads(bead_list[res_type], g_var.cg_residues[res_type][residue], res_type)
        if bead_list[res_type] != sorted(bead_list_cg):
            if len(bead_list[res_type]) == len(bead_list_cg):
                fix_atom_wrap(bead_list[res_type], bead_list_cg, res_type, residue)
            else:
                print('There is a issue with residue: '+res_type+' '+str(residue+1))
                sys.exit('number of atomistic fragments: '+str(len(bead_list[res_type]))+' does not equal number of CG beads: '+str(len(bead_list_cg)))

def sanity_check():
#### runs through every bead and checks whether it exists
    for res_type in g_var.cg_residues:
        if res_type in ['PROTEIN', 'OTHER']:
            sanity_check_protein_other(res_type)
        elif res_type in g_var.sol_residues or res_type in g_var.ion_residues:
            sanity_check_solvent(res_type)
        else:
            sanity_check_non_protein(res_type)

def fix_atom_wrap(bead_list_frag, bead_list_cg, section, resid):
    for bead in bead_list_cg:
        if bead not in bead_list_frag:
            new_bead = bead[1:]+bead[0]
            if new_bead in bead_list_frag and new_bead not in bead_list_cg:
                g_var.cg_residues[section][resid][new_bead] = g_var.cg_residues[section][resid][bead]
                del g_var.cg_residues[section][resid][bead]
            else:
                print('There is a issue with residue: '+section+' '+str(resid+1))
                print('input file list:\n',sorted(bead_list_cg))
                print('\ncannot find: '+bead+' or '+new_bead+' in fragment list:')
                sys.exit(sorted(bead_list_frag))

#####  Sanity check end

#### rotations section

def rotate_atom(coord, center,xyz_rot_apply):
#### rotates atom around center
    coord =  coord-center  #### centers COM coordinates to 0,0,0
    if np.any(xyz_rot_apply): 
        coord=np.array(coord).dot(xyz_rot_apply)
    coord =  coord+center #### translates coord back by original offset
    return coord

def kabsch_rotate(at_connections,cg_connections):
    ## A solution for the best rotation to relate two sets of vectors. By WOLFGANG KABSCH
    ## http://en.wikipedia.org/wiki/Kabsch_algorithm
    Cov_mat = np.dot(np.transpose(at_connections), cg_connections)
    V, S, W = np.linalg.svd(Cov_mat)
    reflection = np.linalg.det(V) * np.linalg.det(W)
    if reflection < 0:
        S[-1] = -S[-1]
        V[:, -1] = -V[:, -1]
    rot_mat = np.dot(V, W)
    return rot_mat

## rotations end
## alignments 

def find_cross_vector(ca):
#### finds cross vector of the atoms CA, CA+1, CA+2 
    AB = ca[0]-ca[1]
    AC = ca[0]-ca[2] 
    cross_vector = np.cross(AB, AC)
    cross_vector = cross_vector/np.linalg.norm(cross_vector)
    return cross_vector

def noramlised_vector(c1, c2):
    initial_vector= c1-c2
    initial_vector = initial_vector/np.linalg.norm(initial_vector)
    return initial_vector


def align_to_vector(v1, v2):
#### returns the rotation matrix to rotate v1 to v2
    v = np.cross(v1,v2)
    c = np.dot(v1,v2)
    s = np.linalg.norm(v)

    rotation=np.array([[   0, -v[2],  v[1]],
                       [v[2],     0, -v[0]],
                       [-v[1], v[0],    0], 
                       ])

    r = np.identity(3) - rotation + np.matmul(rotation,rotation) * ((1 - c)/(s**2))
    return r

def align_at_frag_to_CG_frag(at_com, cg_com, group):
#### aligns atomistic fragment to cg bead
    COM_vector=at_com-cg_com ### gets vector between COM of atoms in fragment and cg bead 
    for bead in group:
        for atom in group[bead]: ### runs through atoms in fragments and centers on the cg bead 
            group[bead][atom]['coord']=group[bead][atom]['coord']-COM_vector
    return group, COM_vector

def COM(mass, fragment):
#### returns center of mass of fragment
    try:
        if np.any(np.array(mass)[:,3]):   
            return np.average(np.array(mass)[:,:3], axis=0, weights=np.array(mass)[:,3])
    except:
        for bead in fragment:
            print(bead, fragment[bead], '\n')
            print(mass)
            sys.exit('missing the mass one of the atoms in '+fragment[bead][1]['res_type'])      

def rigid_fit(group, frag_mass, resid, cg):
#### rigid fits group to CG beads
    rigid_mass_at = []
    rigid_mass_cg = []
    for bead in group:
        rigid_mass_at+=frag_mass[bead]
        rigid_mass_cg.append(cg[bead]['coord'])
    rigid_mass_at = COM(rigid_mass_at, group)
    rigid_mass_cg = np.mean(rigid_mass_cg, axis=0)
    
    group, COM_vector = align_at_frag_to_CG_frag(rigid_mass_at, rigid_mass_cg, group)
    at_frag_centers = {}
    cg_frag_centers = {}
    for bead in group:
        cg_frag_centers[bead] = cg[bead]['coord']
        at_frag_centers[bead] = COM(frag_mass[bead], bead)
        at_frag_centers[bead] -= COM_vector
    rigid_mass_cg, at_frag_centers, cg_frag_centers, group
    return rigid_mass_cg, at_frag_centers, cg_frag_centers, group

## alignment end

## overlap checker

def overlapping_atoms(tree):
    overlapped_ndx = tree.query_ball_tree(tree, r=g_var.args.ov)
    overlapped_cut = [ndx for ndx in overlapped_ndx if len(ndx) >1]      
    overlapped_cut.sort()
    overlapped=list(overlapped_cut for overlapped_cut,_ in itertools.groupby(overlapped_cut))
    return overlapped

def check_atom_overlap(coordinates):
#### creates tree of atom coordinates
    tree = cKDTree(coordinates)
    overlapped = overlapping_atoms(tree)
#### runs through overlapping atoms and moves atom in a random diection until it is no longer overlapping
    while len(overlapped) > 0:
        for ndx_val, ndx in enumerate(overlapped):
            if np.round((ndx_val/len(overlapped))*100,2).is_integer() and len(overlapped) > 30:
                print('fixing '+str(len(overlapped))+' overlapped atoms: '+str(np.round((ndx_val/len(overlapped))*100,2))+' %', end='\r')
            xyz_check = np.array([coordinates[ndx[0]][0]+np.random.uniform(-0.2, 0.2), coordinates[ndx[0]][1]+np.random.uniform(-0.2, 0.2),coordinates[ndx[0]][2]+np.random.uniform(-0.2, 0.2)])
            while len(tree.query_ball_point(xyz_check, r=g_var.args.ov)) > 1:
                xyz_check = np.array([coordinates[ndx[0]][0]+np.random.uniform(-0.2, 0.2), coordinates[ndx[0]][1]+np.random.uniform(-0.2, 0.2),coordinates[ndx[0]][2]+np.random.uniform(-0.2, 0.2)])
            coordinates[ndx[0]]=xyz_check
            tree = cKDTree(coordinates)
        if len(overlapped) > 30:
            print('{:<100}'.format(''), end='\r')
        overlapped = overlapping_atoms(tree)
    return coordinates

## overlap end
## get fragment information

def split_fragment_names(line, residue, resname):
    bead = gen.strip_header(line)
    group = g_var.res_top[resname]['GROUPS'][bead]
    if group not in residue:
        residue[group] = {}
    if bead not in residue[group]:
        residue[group][bead]={} 
    return residue, group, bead   

def get_atomistic(frag_location, resname=False):
    if not resname:
        resname = frag_location.split('/')[-1][:-4]
#### read in atomistic fragments into dictionary    
    residue = {} ## a dictionary of bead in each residue eg residue[group][bead][atom number(1)][residue_name(ASP)/coordinates(coord)/atom name(C)/connectivity(2)/atom_mass(12)]
    fragment_mass = {}
    with open(frag_location, 'r') as pdb_input:
        for line_nr, line in enumerate(pdb_input.readlines()):
            if line.startswith('['):
                residue, group, bead = split_fragment_names(line, residue, resname)
                fragment_mass[bead]=[]
            if line.startswith('ATOM'):
                line_sep = gen.pdbatom(line) ## splits up pdb line
                residue[group][bead][line_sep['atom_number']]={'coord':np.array([line_sep['x']*g_var.sf,line_sep['y']*g_var.sf,line_sep['z']*g_var.sf]),
                                                                'atom':line_sep['atom_name'],'resid':1, 'resid_ori':line_sep['residue_id'],'res_type':line_sep['residue_name'],
                                                                'frag_mass':1}    
#### updates fragment mass   
                if not gen.is_hydrogen(line_sep['atom_name']):
                    if line_sep['atom_name'] in g_var.res_top[resname]['atom_masses']:
                        residue[group][bead][line_sep['atom_number']]['frag_mass']=g_var.res_top[resname]['atom_masses'][line_sep['atom_name']]  ### updates atom masses with crude approximations
                        fragment_mass[bead].append([line_sep['x']*g_var.sf,line_sep['y']*g_var.sf,line_sep['z']*g_var.sf,g_var.res_top[resname]['atom_masses'][line_sep['atom_name']]])               
                else:
                    fragment_mass[bead].append([line_sep['x']*g_var.sf,line_sep['y']*g_var.sf,line_sep['z']*g_var.sf,1])
    return residue, fragment_mass

def connectivity(cg, at_frag_centers, cg_frag_centers, group, group_number):
#### returns the connections between the atomistic and coarse grain beads
    at_connection, cg_connection=[],[]
    for bead in cg:
        resname = cg[bead]['residue_name']
        break
    if resname in g_var.sorted_connect:
        if len(g_var.sorted_connect[resname]) > 0:
            for group_bead in group: 
                for bead_atom in group[group_bead]:
                    if bead_atom in g_var.sorted_connect[resname][int(group_number)]:
                        skip = False
                        for bead_connect in g_var.sorted_connect[resname][int(group_number)][bead_atom]:
                            if 'cg_connect' not in locals():
                                if bead_connect in cg:
                                    cg_connect = [cg[bead_connect]['coord']]
                                else:
                                    skip = True
                            else:
                                if bead_connect in cg:
                                    cg_connect.append(cg[bead_connect]['coord'])
                        if not skip:
                            at_connection.append(group[group_bead][bead_atom]['coord'])
                            cg_connection.append(np.mean(np.array(cg_connect), axis=0))    
    if len(at_frag_centers)  > 1:         
        for bead in at_frag_centers:
            cg_connection.append(cg_frag_centers[bead])
            at_connection.append(at_frag_centers[bead])   
    return at_connection, cg_connection    

def get_rotation(cg_connect, at_connect, center, resname, group, cg_resid):
    if len(at_connect) == len(cg_connect):
        if len(cg_connect) == 0:
            return gen.AnglesToRotMat([np.random.uniform(0, math.pi*2), np.random.uniform(0, math.pi*2), np.random.uniform(0, math.pi*2)])
        else:
            return kabsch_rotate(np.array(at_connect)-center, np.array(cg_connect)-center)
    else:
        print('atom connections: '+str(len(at_connect))+' does not match CG connections: '+str(len(cg_connect)))
        sys.exit('residue number: '+str(cg_resid)+', residue type: '+str(resname)+', group: '+group)

def apply_rotations(atomistic_fragments,cg_resid, group_fit, center, xyz_rot_apply):
    for bead in group_fit:
        for atom in group_fit[bead]:
            group_fit[bead][atom]['coord'] = rotate_atom(group_fit[bead][atom]['coord'], center, xyz_rot_apply)   
            atomistic_fragments[cg_resid][atom] = group_fit[bead][atom].copy()
    return atomistic_fragments

### end fragment information
### linked residue connectivity start

def BB_connectivity(at_connections,cg_connections, cg_residues, at_residues, residue_number, BB_bead, resname):
    con_atoms = {}
    for atom in at_residues:
        # resname = at_residues[atom]['res_type']
        if at_residues[atom]['atom'] in g_var.res_top[resname]['CONNECT'][BB_bead]['atom']:
            con_atoms[g_var.res_top[resname]['CONNECT'][BB_bead]['atom'].index(at_residues[atom]['atom'])]=atom
    new_chain=False
    for con in con_atoms:
        con_resid = residue_number+g_var.res_top[resname]['CONNECT'][BB_bead]['dir'][con]
        if con_resid in cg_residues:
            xyz_cur = cg_residues[residue_number][BB_bead]['coord']
            if g_var.res_top[resname]['CONNECT'][BB_bead]['Con_Bd'][con] in cg_residues[con_resid]:
                xyz_con = cg_residues[con_resid][g_var.res_top[resname]['CONNECT'][BB_bead]['Con_Bd'][con]]['coord']
                if gen.calculate_distance(xyz_con, xyz_cur) < 6:
                    cg_connections.append(xyz_con)
                    at_connections.append(at_residues[con_atoms[con]]['coord'])
                else:
                    if g_var.res_top[resname]['CONNECT'][BB_bead]['dir'][con] > 0:
                        new_chain=True
            else:
                    if g_var.res_top[resname]['CONNECT'][BB_bead]['dir'][con] > 0:
                        new_chain=True
        else:
            if g_var.res_top[resname]['CONNECT'][BB_bead]['dir'][con] > 0:
                new_chain=True
    return at_connections,cg_connections, new_chain

################################################################### Merged system

def merge_indivdual_chain_pdbs(file, end, res_type):
#### reads in each chain into merge list
    merge, merged_coords = [],[]
    count = 0
    restraint_count = -1
    for chain in range(0,g_var.system[res_type]):
        merge_temp = []
        if os.path.exists(file+'_'+str(chain)+end):
            with open(file+'_'+str(chain)+end, 'r') as pdb_input:
                merge_temp += read_in.filter_input(pdb_input.readlines(), False)
        else:
            sys.exit('cannot find chain: '+file+'_'+str(chain)+end)
        if res_type+'_aligned' in file:  
            count, restraint_count = at_mod_p.create_disres(merge_temp, chain, file, count, restraint_count)
        merge, merge_coords = fix_chirality(merge,merge_temp,merged_coords, res_type)   
    if res_type+'_aligned' not in file:
        coords, index_conversion = index_conversion_generate(merge, merge_coords)
    if 'aligned' in file.split('/')[-1]:
        write_pdb(merge, merge_coords, {}, g_var.working_dir+'PROTEIN/PROTEIN_aligned_merged.pdb')
    else:
        write_pdb(merge, coords,index_conversion, g_var.working_dir+res_type+'/'+res_type+'_de_novo_merged.pdb')

def index_conversion_generate(merge, merge_coords):
    index_conversion = {}
    coords = []
    for at_val, atom in enumerate(merge):
        if not atom['atom_name'].startswith('M'):
            index_conversion[at_val] = len(coords)
            coords.append(merge_coords[at_val])
    coords = check_atom_overlap(coords)
    return coords, index_conversion

def write_pdb(merge, merge_coords, index_conversion, write_file):
#### creates merged pdb and writes chains to it
    if not os.path.exists(write_file):
        pdb_output=gen.create_pdb(write_file)
        atom_counter=0
        for line_val, line in enumerate(merge):
            if line_val in index_conversion:
                x, y, z = gen.trunc_coord(merge_coords[index_conversion[line_val]])
            else:
                if 'coord' in line:
                    x, y, z = gen.trunc_coord(line['coord'])
                else:
                    x, y, z = gen.trunc_coord([line['x'],line['y'],line['z']])
            if atom_counter >= 99_999:
                atom_counter=1
            else:
                atom_counter+=1      
            pdb_output.write(g_var.pdbline%((atom_counter, line['atom_name'], line['residue_name'],' ',line['residue_id'],\
            x,y,z,1,0))+'\n')
        pdb_output.close()

def merge_system_pdbs(protein):
    os.chdir(g_var.merged_directory)
#### create merged pdb 
    if not os.path.exists(g_var.merged_directory+'merged_cg2at'+protein+'.pdb'):
        merge=[]
        merge_coords=[]
    #### run through every residue type in cg_residues
        done = []
        if 'novo' in protein:
            print('checking for atom overlap in : '+protein[1:])
        for segment, residue_type in enumerate(g_var.system):
            if residue_type not in done:
                if residue_type not in ['PROTEIN', 'OTHER']:
                        start_num = len(merge_coords)
                        merge, merge_coords = read_in_merged_pdbs(merge, merge_coords, 
                                                                g_var.working_dir+residue_type+'/'+residue_type+'_merged.pdb')
                        g_var.np_blocks[residue_type]=[start_num,len(merge_coords)]
                elif residue_type == 'OTHER':
                    merge, merge_coords = read_in_merged_pdbs(merge, merge_coords, 
                                                            g_var.working_dir+residue_type+'/'+residue_type+'_de_novo_merged.pdb')
                elif residue_type == 'PROTEIN':
                    merge, merge_coords = read_in_merged_pdbs(merge, merge_coords, 
                                                            g_var.working_dir+residue_type+'/'+residue_type+protein+'_merged.pdb')     
                done.append(residue_type)
        index_conversion = {}
        if 'novo' in protein:
            coords, index_conversion = index_conversion_generate(merge, merge_coords)
            write_pdb(merge, coords, index_conversion, g_var.merged_directory+'merged_cg2at'+protein+'.pdb')
        else:
            write_pdb(merge, merge_coords, {}, g_var.merged_directory+'merged_cg2at'+protein+'.pdb')

def read_in_merged_pdbs(merge, merge_coords, location):
    if os.path.exists(location):
    #### opens pdb files and writes straight to merged_cg2at pdb
        with open(location, 'r') as pdb_input:
            read_in_atoms = read_in.filter_input(pdb_input.readlines(), False)
            merge_coords += [[line_sep['x'],line_sep['y'],line_sep['z']] for line_sep in read_in_atoms ]
            return merge+read_in_atoms, merge_coords
    else:
        sys.exit('cannot find minimised residue: \n'+ location) 

def check_overlap_chain(chain, input,res_type):
    if not os.path.exists(g_var.working_dir+res_type+'/'+res_type+'_'+input+str(chain)+'_gmx_checked.pdb'):
        lines, coords = read_in_merged_pdbs([], [], res_type+'_'+input+str(chain)+'_gmx.pdb')
        if res_type == 'PROTEIN':
            coords = at_mod_p.correct_amide_h(lines, coords)

        coords, index_conversion = index_conversion_generate(lines, coords)
        write_pdb(lines, coords, index_conversion, g_var.working_dir+res_type+'/'+res_type+'_'+input+str(chain)+'_gmx_checked.pdb')

## fix chirality errors

def fetch_chiral_coord(merge_temp, residue_type):
    chiral_atoms={}
    coord=[]
    for atom in range(len(merge_temp)):
        if residue_type in ['PROTEIN', 'OTHER']:
            resname = gen.check_alternate_resname(merge_temp[atom]['residue_name'])
        else:
            resname = gen.check_alternate_resname(residue_type)
        if len(g_var.res_top[resname]['CHIRAL']['atoms']) > 0:
            if merge_temp[atom]['atom_name'] in g_var.res_top[resname]['CHIRAL']['atoms']:
                if merge_temp[atom]['residue_id'] not in chiral_atoms:
                    chiral_atoms[merge_temp[atom]['residue_id']]={}
                chiral_atoms[merge_temp[atom]['residue_id']][merge_temp[atom]['atom_name']]=atom
        coord.append(np.array([merge_temp[atom]['x'],merge_temp[atom]['y'],merge_temp[atom]['z']]))
    return chiral_atoms, coord


def get_atom_move(merge_temp, resname, residue, chiral_group, chiral_atoms):
    stat = merge_temp[chiral_atoms[residue][chiral_group]].copy()
    atom_move = {'stat':np.array([stat['x'],stat['y'],stat['z']]), 'm':'', 'c1':'', 'c2':'', 'c3':''}
    for chir_atom in atom_move:
        if chir_atom != 'stat':
            test = merge_temp[chiral_atoms[residue][g_var.res_top[resname]['CHIRAL'][chiral_group][chir_atom]]].copy()
            atom_move[chir_atom]= np.array([test['x'],test['y'],test['z']])
            if gen.calculate_distance(atom_move['stat'], atom_move[chir_atom]) > 10:
                atom_move[chir_atom] = np.array(read_in.brute_mic(atom_move['stat'],atom_move[chir_atom]))
    return atom_move


def fix_chirality(merge, merge_temp, merged_coords, residue_type):
#### fixes chiral groups
    chiral_atoms, coord = fetch_chiral_coord(merge_temp, residue_type)
    for residue in chiral_atoms:
        if residue_type in ['PROTEIN', 'OTHER']:
            for atom in chiral_atoms[residue]:
                resname = merge_temp[chiral_atoms[residue][atom]]['residue_name']
                resname = gen.check_alternate_resname(resname)
                break
        else:
            resname = gen.check_alternate_resname(residue_type)

        for chiral_group in g_var.res_top[resname]['CHIRAL']:
            if chiral_group != 'atoms':
                atom_move = get_atom_move(merge_temp, resname, residue, chiral_group, chiral_atoms)

                S_M = atom_move['m'] - atom_move['stat']
                rotation = align_to_vector(S_M, [0,0,1])
                c1_coord = (atom_move['c1'] - atom_move['stat']).dot(rotation)
                c2_coord = (atom_move['c2'] - atom_move['stat']).dot(rotation)
                c3_coord = (atom_move['c3'] - atom_move['stat']).dot(rotation)

                if gen.angle_clockwise(c1_coord[0:2], c2_coord[0:2]) > gen.angle_clockwise(c1_coord[0:2], c3_coord[0:2]):
                    for ax_val, axis in enumerate(['x', 'y', 'z']):
                        merge_temp[chiral_atoms[residue][g_var.res_top[resname]['CHIRAL'][chiral_group]['m']]][axis] = merge_temp[chiral_atoms[residue][g_var.res_top[resname]['CHIRAL'][chiral_group]['m']]][axis] - (3*S_M[ax_val])   
                        merge_temp[chiral_atoms[residue][chiral_group]][axis] = merge_temp[chiral_atoms[residue][chiral_group]][axis] - (S_M[ax_val])        
                    coord[chiral_atoms[residue][g_var.res_top[resname]['CHIRAL'][chiral_group]['m']]] -=  (2*S_M) #move_coord -
                    coord[chiral_atoms[residue][chiral_group]] -=  (0.25*S_M) 
    merge+=merge_temp
    merged_coords+=coord
    return merge, merged_coords

def get_chiral_non_carbonyl(at):
    chiral = {}
    for atom in at:
        if at[atom]['atom'] in g_var.res_top[at[atom]['res_type']]['CHIRAL']['atoms']:
            chiral[at[atom]['atom']] = atom
    return chiral

def correct_chiral_atoms(at, chiral):
    for chiral_group in g_var.res_top[at[1]['res_type']]['CHIRAL']:
        if chiral_group != 'atoms':            
            atom_list = [chiral[chiral_group]]
            for chiral_atoms in ['c1','c2','c3']:
                atom_list.append(chiral[g_var.res_top[at[1]['res_type']]['CHIRAL'][chiral_group][chiral_atoms]])
            cross_vector_chiral = find_cross_vector( [at[atom_list[3]]['coord'], at[atom_list[2]]['coord'], at[atom_list[1]]['coord']])
            norm_vector_chiral = noramlised_vector(at[chiral[g_var.res_top[at[1]['res_type']]['CHIRAL'][chiral_group]['m']]]['coord'], at[atom_list[0]]['coord'])
            if not np.dot(cross_vector_chiral, norm_vector_chiral) > 0.5:
                if g_var.res_top[at[1]['res_type']]['CHIRAL'][chiral_group]['m'] in chiral:
                    at[atom_list[0]]['coord'] = at[atom_list[0]]['coord'] + cross_vector_chiral*0.5
                    m = chiral[g_var.res_top[at[1]['res_type']]['CHIRAL'][chiral_group]['m']]
                    at[m]['coord'] = at[m]['coord'] + cross_vector_chiral*2
    return at

## chirality end
## hydrogen orientaion checker

def check_hydrogens(residue):
#### finds the connecting carbons and their associated carbons [carbon atom, hydrogen ref number, connecting ref number]    
    for atom_num, atom in enumerate(residue):
        resname=residue[atom]['res_type']
        break
    for atom in g_var.hydrogen[resname]:
        h_coord = []
        for group in g_var.sorted_connect[resname]:
            if atom in g_var.sorted_connect[resname][group]:
                for hydrogen in g_var.hydrogen[resname][atom]:
                    h_coord.append(residue[hydrogen]['coord'])
                h_com=np.mean(np.array(h_coord), axis=0)
                for heavy_bond in g_var.heavy_bond[resname][atom]:
                    for group_check in g_var.sorted_connect[resname]:
                        if heavy_bond in g_var.sorted_connect[resname][group_check] and group_check != group:
                            skip = False
                            if heavy_bond in residue:
                                con_heavy_atom = heavy_bond
                                con_heavy_atom_co =  residue[con_heavy_atom]['coord']
                            else:
                                skip = True
                if not skip:
                #### vector between H COM and bonded carbon 
                    vector=np.array([h_com[0]-residue[atom]['coord'][0],h_com[1]-residue[atom]['coord'][1],h_com[2]-residue[atom]['coord'][2]])
                    h_com_f=h_com+vector*2
                    d1 = gen.calculate_distance(h_com, con_heavy_atom_co)    
                    d2 = gen.calculate_distance(h_com_f, con_heavy_atom_co)   
                    if d2 < d1:
                        for h_at in g_var.hydrogen[resname][atom]:
                            residue[h_at]['coord']=residue[h_at]['coord']-vector*2
    return residue

## check for threaded lipids
def check_ringed_lipids(protein):
    print('Checking for ringed lipids')
    if not os.path.exists(g_var.merged_directory+'checked_ringed_lipid_de_novo.pdb'):
        if not os.path.exists(g_var.merged_directory+'merged_cg2at_threaded.pdb'):
            os.chdir(g_var.merged_directory)
            merge, merge_coords = read_in_merged_pdbs([], [], protein)
            ringed=False
            lipid_atoms = []
            with open(g_var.merged_directory+'threaded_lipids.dat', 'w') as ring_ouput:
                for at_val, atom in enumerate(merge): 
                    resname = get_np_resname(at_val)
                    if resname in g_var.np_residues:
                        offset = fetch_start_of_residue_np(at_val, resname)
                        if atom['atom_number']-offset in g_var.heavy_bond[resname]:
                            for at_bond in g_var.heavy_bond[resname][atom['atom_number']-offset]:
                                at_bond -=1
                                if merge[at_bond+offset]['atom_number'] > merge[at_val]['atom_number']:
                                    merge[at_bond+offset]['x'], merge[at_bond+offset]['y'], merge[at_bond+offset]['z'] = np.array(read_in.brute_mic(merge_coords[at_val],merge_coords[at_bond+offset]))
                                    merge_coords[at_bond+offset] = merge[at_bond+offset]['x'], merge[at_bond+offset]['y'], merge[at_bond+offset]['z']
                                    dist = gen.calculate_distance(merge_coords[at_val], merge_coords[at_bond+offset])
                                    if 2 < dist < 6:
                                        lipid_atoms.append([at_val, at_bond+offset, (np.array(merge_coords[at_val])+np.array(merge_coords[at_bond+offset]))/2])
                                        ring_ouput.write('{0:6}{1:6}{2:2}{3:4}{4:2}{5:7}{6:5}{7:5}{8:5}{9:5}{10:5}{11:5}\n'.format(
                                                            'distance: ',str(np.round(dist,2)),'residue: ', merge[at_val]['residue_name'], merge[at_val]['residue_id'],
                                                            ' atom_1: ', merge[at_val]['atom_name'],
                                                            'atom_2: ', merge[at_bond+offset]['atom_name'], 'rough line num: ', at_val, at_bond+offset))
                                        ringed = True
        if ringed or os.path.exists(g_var.merged_directory+'merged_cg2at_threaded.pdb'):
            print('Found '+str(len(lipid_atoms))+' abnormal bonds, now attempting to fix.')
            print('See this file for a complete list: '+g_var.merged_directory+'threaded_lipids.dat')
            fix_threaded_lipids(lipid_atoms, merge, merge_coords)
        else:
            gen.file_copy_and_check(g_var.merged_directory+'MIN/merged_cg2at_de_novo_minimised.pdb', g_var.merged_directory+'checked_ringed_lipid_de_novo.pdb')

def fetch_start_of_residue(at, merge):
    count = at
    while True:
        if merge[count]['residue_id'] != merge[at]['residue_id']:
            count+=1
            return count
        else:
            count-=1

def fetch_start_of_residue_np(atom, resname):
            prev=g_var.np_blocks[resname][0]
            range_np_block = g_var.np_blocks[resname][1]-g_var.np_blocks[resname][0]
            for range_np_block_ind in range(g_var.np_blocks[resname][0],g_var.np_blocks[resname][1]+1, int(range_np_block/g_var.system[resname])):
                if prev <= atom < range_np_block_ind:
                    return prev
                prev = range_np_block_ind

def get_np_resname(atom):
    for res_check in g_var.np_blocks:
        if g_var.np_blocks[res_check][0] <= atom < g_var.np_blocks[res_check][1]:
            return res_check 

def fix_threaded_lipids(lipid_atoms, merge, merge_coords):
    if not os.path.exists(g_var.merged_directory+'merged_cg2at_threaded.pdb'):
        tree = cKDTree(merge_coords)         
        for threaded in lipid_atoms:  
            resname = get_np_resname(threaded[0]) 
            
            atoms = tree.query_ball_point(threaded[2], r=3)
            for at in atoms:
                if merge[at]['residue_id'] != merge[threaded[0]]['residue_id']:
                    P_count = fetch_start_of_residue(at, merge)
                    break

            NP_count = fetch_start_of_residue_np(threaded[0], resname)
            bb = []
            if 'P_count' not in locals():
                sys.exit('There is an issue with the bond length detection')
            for at in merge[P_count:]:
                if at['residue_id'] != merge[P_count]['residue_id']:
                    break
                if at['atom_name'] in g_var.res_top[at['residue_name']]['ATOMS']:
                    bb.append([at['x'], at['y'], at['z']])
            bb = np.mean(np.array(bb), axis=0)
            BB_M3 = (threaded[2]-bb)/np.linalg.norm((threaded[2]-bb))      
            
            for heavy_atom in threaded[:2]:
                merge_coords[heavy_atom] += BB_M3*3 
                merge[heavy_atom]['x'], merge[heavy_atom]['y'], merge[heavy_atom]['z'] = merge_coords[heavy_atom]
                for hydrogen in g_var.hydrogen[resname][heavy_atom-NP_count+1]:
                    merge_coords[NP_count+hydrogen-1] += BB_M3*3 
                    merge[NP_count+hydrogen-1]['x'], merge[NP_count+hydrogen-1]['y'], merge[NP_count+hydrogen-1]['z'] = merge_coords[NP_count+hydrogen-1]

        coords, index_conversion = index_conversion_generate(merge, merge_coords)
        write_pdb(merge, coords, index_conversion, g_var.merged_directory+'merged_cg2at_threaded.pdb')

    if not os.path.exists(g_var.merged_directory+'MIN/merged_cg2at_threaded_minimised.pdb'):
        gro.minimise_merged_pdbs('_threaded')
    gen.file_copy_and_check(g_var.merged_directory+'MIN/merged_cg2at_threaded_minimised.pdb', g_var.merged_directory+'checked_ringed_lipid_de_novo.pdb')



#!/usr/bin/env python3

import os, sys
import numpy as np
import math
import gen, g_var, at_mod



def build_atomistic_system(residue_type):
    system={}
#### for each residue type covert to atomistic except protein
    print('Converting residue type: ' +residue_type)
#### creates folder for residue type
    gen.mkdir_directory(g_var.working_dir+residue_type)
    if not os.path.exists(g_var.working_dir+residue_type+'/'+residue_type+'_merged.pdb'):
        atomistic_fragments, system[residue_type] = at_np_solvent(residue_type, g_var.cg_residues[residue_type])
        write_solvent(atomistic_fragments, residue_type)
    elif residue_type in g_var.sol_residues:
        atomistic_fragments, system[residue_type] = read_solvent_conversion(residue_type, g_var.cg_residues[residue_type])
    else:
        system[residue_type]=len(g_var.cg_residues[residue_type])
            
    print('{:<100}'.format(''), end='\r')
    print('Finished converting: '+residue_type)
    return system 

def write_solvent(atomistic_fragments, residue_type):
    merge_coord = [line['coord'] for line in atomistic_fragments]
    coord, index_conversion = at_mod.index_conversion_generate(atomistic_fragments, merge_coord)
    at_mod.write_pdb(atomistic_fragments, coord, index_conversion, g_var.working_dir+residue_type+'/'+residue_type+'_all.pdb')

def read_solvent_conversion(cg_residue_type,cg_residues):
    residue_type={}
    residue_type_mass={}
    for cg_resid, cg_residue in enumerate(cg_residues):
        frag_location=gen.fragment_location(cg_residue_type) ### get fragment location from database
        residue_type[cg_residue_type], residue_type_mass[cg_residue_type] = at_mod.get_atomistic(frag_location, cg_residue_type)

        for group in residue_type[cg_residue_type]:
            sol_p_bead = 0
            for frag in residue_type[cg_residue_type][group].values():
                for atom in frag.values():
                    if int(atom['resid_ori']) > sol_p_bead:
                        sol_p_bead = int(atom['resid_ori'])
            return sol_p_bead, sol_p_bead*len(cg_residues)
    
    sys.exit('There is an issue with the solvent recalculation')

def at_np_solvent(cg_residue_type,cg_residues):   
    atomistic_fragments={}  #### residue dictionary
#### run through every residue in a particular residue type
    residue_type={}
    residue_type_mass={}
    atomistic_fragments_list = []
    for cg_resid, cg_residue in enumerate(cg_residues):
        atomistic_fragments[cg_resid]={}
        frag_location=gen.fragment_location(cg_residue_type) ### get fragment location from database
        residue_type[cg_residue_type], residue_type_mass[cg_residue_type] = at_mod.get_atomistic(frag_location, cg_residue_type)
        for group in residue_type[cg_residue_type]:
            center, at_frag_centers, cg_frag_centers, group_fit = at_mod.rigid_fit(residue_type[cg_residue_type][group], 
                                                                                    residue_type_mass[cg_residue_type], 
                                                                                    cg_residue, cg_residues[cg_residue])
            at_connect, cg_connect = at_mod.connectivity(cg_residues[cg_residue], at_frag_centers, cg_frag_centers, group_fit, group)
            xyz_rot_apply=at_mod.get_rotation(cg_connect, at_connect, center, cg_residue_type, group, cg_resid)

            atomistic_fragments = at_mod.apply_rotations(atomistic_fragments,cg_resid, group_fit, center, xyz_rot_apply)
            
        if cg_residue_type in g_var.np_residues:
            atomistic_fragments[cg_resid] = at_mod.check_hydrogens(atomistic_fragments[cg_resid])
        if len(g_var.res_top[cg_residue_type]['CHIRAL']) > 1:
            chiral = at_mod.get_chiral_non_carbonyl(atomistic_fragments[cg_resid])
            atomistic_fragments[cg_resid] = at_mod.correct_chiral_atoms(atomistic_fragments[cg_resid], chiral)            
        atomistic_fragments_list, sol_p_bead =  sort_np_dictionary(atomistic_fragments[cg_resid], atomistic_fragments_list)
    if cg_residue_type in g_var.sol_residues:
        return atomistic_fragments_list, sol_p_bead*len(cg_residues)
    else:
        return atomistic_fragments_list, len(atomistic_fragments)

def sort_np_dictionary(atomistic_fragments, atomistic_fragments_list):
    sol_p_bead = 0
    dict_start = np.min(list(atomistic_fragments.keys()))
    for atom_val in range(dict_start, len(atomistic_fragments)+dict_start):
        atom = atomistic_fragments[atom_val]
        atom['atom_name'] = atom.pop('atom')
        atom['residue_name'] = atom.pop('res_type')
        atom['residue_id'] = atom.pop('resid')
        atomistic_fragments_list.append(atom)    
        if atom['resid_ori'] > sol_p_bead:
            sol_p_bead = atom['resid_ori']
    return atomistic_fragments_list, sol_p_bead 



#!/usr/bin/env python3

import sys, os
import numpy as np
import math
import copy
import gen, g_var



def read_initial_cg_pdb(test=False):
    if not test:
        print('\nThis script is now hopefully doing the following (Good luck):\n')
    residue_list={} ## a dictionary of bead in each residue eg residue_list[bead name(BB)][residue_name(PO4)/coordinates(coord)]
    count=0  ### residue counter initialisation
    with open(g_var.input_directory+'CG_INPUT.pdb', 'r') as pdb_input:
        pdb_lines_atoms, box_vec = filter_input(pdb_input.readlines())
        for line_val, line_sep in enumerate(pdb_lines_atoms):
            if np.round((line_val/len(pdb_lines_atoms))*100,2).is_integer() and not test:
                print('Reading in your CG representation: ',np.round((line_val/len(pdb_lines_atoms))*100,2),'%', end='\r')
            line_sep = update_residue_names(line_sep)
            line_sep['atom_name'], line_sep['residue_name'] = swap(line_sep['atom_name'], line_sep['residue_name'], line_sep['residue_id']) ## implements swap group
            if 'SKIP' not in [line_sep['atom_name'].upper(), line_sep['residue_name'].upper()]:
#### set up resnames in dictionaries
                add_residue_to_dictionary(line_sep)
#### sets up previous resid id 
                if 'residue_prev' not in locals(): 
                    residue_prev=line_sep.copy() 
#### if resid the same as previous line
                if residue_prev['residue_id'] == line_sep['residue_id'] and line_sep['residue_name'] == residue_prev['residue_name'] \
                                                                         and line_sep['atom_name'] not in residue_list:   ### if resid is the same as the previous line, it adds resname and coordinates to the atom name key in residue_list 
                    residue_list[line_sep['atom_name']]={'residue_name':line_sep['residue_name'],
                                                        'coord':np.array([line_sep['x'],line_sep['y'],line_sep['z']])}
                    line_sep_prev=line_sep.copy()
                else: 
                    add_to_cg_database(line_sep_prev, count, residue_list)
#### updates dictionaries and counters
                    residue_list={}  ### resets residue list
                    count+=1 ### moves counter along to next residue
                    residue_list[line_sep['atom_name']]={'residue_name':line_sep['residue_name'],
                                                        'coord':np.array([line_sep['x'],line_sep['y'],line_sep['z']])} ### it adds resname and coordinates to the atom name key in residue_list
                    residue_prev=line_sep.copy()    ### updates residue_prev with new resid
                    line_sep_prev=line_sep.copy()
                if line_val+1 == len(pdb_lines_atoms):
                        residue_list[line_sep['atom_name']]={'residue_name':line_sep['residue_name'],
                                                            'coord':np.array([line_sep['x'],line_sep['y'],line_sep['z']])}
                        add_to_cg_database(line_sep, count, residue_list)
    for key in g_var.cg_residues:
        if len(g_var.cg_residues[key]) == 0:
            sys.exit('there is a issue with the residue type: '+key)
    return box_vec[0]

def filter_input(pdb_lines_raw, CG=True):
    pdb_lines_atoms = [gen.pdbatom(j) for j in pdb_lines_raw if j.startswith('ATOM ')] 
    if len(pdb_lines_atoms) == 0:
        sys.exit('input coarse-grained structure seems to contain no beads')
    if CG:
        box_vec =  [j for j in pdb_lines_raw if j.startswith('CRYST')]
        if len(box_vec) == 0:
            sys.exit('The input file is missing the Box vectors')
        return pdb_lines_atoms, box_vec
    return pdb_lines_atoms


def add_to_cg_database(line_sep_prev, count, residue_list):
    if line_sep_prev['residue_name'] in g_var.sol_residues+g_var.ion_residues+g_var.np_residues :
        g_var.cg_residues[line_sep_prev['residue_name']][count]={} ### then create sub dictionary cg_residues[resname][count]
        g_var.cg_residues[line_sep_prev['residue_name']][count]=residue_list ### adds residue list to dictionary key cg_residues[resname][count]
        if line_sep_prev['residue_name'] in g_var.hydration:
            sol_res_list={}
            sol_res_list[g_var.hydration[line_sep_prev['residue_name']]]=residue_list[line_sep_prev['atom_name']].copy()
            sol_res_list[g_var.hydration[line_sep_prev['residue_name']]]['residue_name']=g_var.hydration[line_sep_prev['residue_name']]
            g_var.cg_residues[g_var.hydration[line_sep_prev['residue_name']]][count]=sol_res_list
    elif line_sep_prev['residue_name'] in g_var.o_residues:
        g_var.cg_residues['OTHER'][count]={} ### then create sub dictionary cg_residues['PROTEIN'][count]
        g_var.cg_residues['OTHER'][count]=residue_list ### adds residue list to dictionary key cg_residues['PROTEIN'][count]
    else:
        g_var.cg_residues['PROTEIN'][count]={} ### then create sub dictionary cg_residues['PROTEIN'][count]
        g_var.cg_residues['PROTEIN'][count]=residue_list ### adds residue list to dictionary key cg_residues['PROTEIN'][count]


def update_residue_names(line_sep):
    if line_sep['residue_name'] =='ION':
        line_sep['residue_name'] = line_sep['atom_name']
    if line_sep['residue_name'] in g_var.alt_res_name:
        line_sep['residue_name'] = g_var.alt_res_name[line_sep['residue_name']]
    if line_sep['residue_name'] in g_var.sol_residues+g_var.ion_residues and line_sep['atom_name'] in g_var.alt_res_name:
        line_sep['atom_name'] = g_var.alt_res_name[line_sep['atom_name']]
    return line_sep

def add_residue_to_dictionary(line_sep):
    if line_sep['residue_name'] in g_var.p_residues: ## if in protein database 
        if 'PROTEIN' not in g_var.cg_residues:  ## if protein does not exist add to dict
            g_var.cg_residues['PROTEIN']={}
    elif line_sep['residue_name'] in g_var.o_residues: ## if in protein database 
        if 'OTHER' not in g_var.cg_residues:  ## if protein does not exist add to dict
            g_var.cg_residues['OTHER']={}
    elif line_sep['residue_name'] in g_var.sol_residues: 
        if line_sep['residue_name'] not in g_var.cg_residues: ## if residue type does not exist add to dict
            g_var.cg_residues[line_sep['residue_name']]={}
    elif line_sep['residue_name'] in g_var.ion_residues: 
        if line_sep['residue_name'] not in g_var.cg_residues: ## if residue type does not exist add to dict
            g_var.cg_residues[line_sep['residue_name']]={}
        if line_sep['residue_name'] in g_var.hydration:
            if g_var.hydration[line_sep['residue_name']] not in g_var.cg_residues:
                g_var.cg_residues[g_var.hydration[line_sep['residue_name']]] = {}
    elif line_sep['residue_name'] in g_var.np_residues:
        if line_sep['residue_name'] not in g_var.cg_residues:
            g_var.cg_residues[line_sep['residue_name']]={}
    elif line_sep['residue_name'] == 'SKIP':
        pass
    else:
        sys.exit('\n'+line_sep['residue_name']+' is not in the fragment database!') 


def check_new_box(coord,box, new_box):
    for xyz_val, xyz in enumerate(new_box):
        if g_var.args.box[xyz_val] != 0:
            lower, upper = (float(box[xyz_val])/2)-(float(xyz)/2), (float(box[xyz_val])/2)+(float(xyz)/2)
            if  lower >= coord[xyz_val] or coord[xyz_val] >= upper: 
                return True
    return False

def real_box_vectors(box_vec):
    x, y, z, yz, xz, xy = [float(i) for i in box_vec.split()[1:7]]
    # return real-space vector
    cyz, cxz, cxy, sxy = math.cos(np.radians(yz)), math.cos(np.radians(xz)), math.cos(np.radians(xy)) , math.sin(np.radians(xy))
    wx, wy             = z*cxz, z*(cyz-cxz*cxy)/sxy
    wz                 = math.sqrt(z**2 - wx**2 - wy**2)
    g_var.r_b_vec = np.array([[x, 0, 0], [y*cxy, y*sxy, 0], [wx, wy, wz]]).T
    g_var.r_b_inv = np.linalg.inv(g_var.r_b_vec).T

def brute_mic(p1, p2):
    result = None
    n = 2
    if gen.calculate_distance(p1, p2) > 10:
        for x in range(-n, n+1):
            for y in range(-n, n+1):
                for z in range(-n, n+1):
                    rp = p2+np.dot(g_var.r_b_vec, [x,y,z])
                    d = gen.calculate_distance(p1, rp)
                    if (result is None) or (result[1] > d):
                        result = (rp, d)
        if result[1] < 10:
            return result[0]
        else:
            return p2
    else:
        return p2

def fix_pbc(box_vec, new_box, box_shift):
#### fixes box PBC
    new_box = new_box.split()[1:4]
    BB_pre_resid = 0
    for residue_type in g_var.cg_residues:
        cut_keys=[]
        print('{:<100}'.format(''), end='\r')
        for res_val, residue in enumerate(g_var.cg_residues[residue_type]):
            if np.round((res_val/len(g_var.cg_residues[residue_type]))*100,2).is_integer():
                print('Fixing PBC of residue type '+residue_type+': ',np.round((res_val/len(g_var.cg_residues[residue_type]))*100,2),'%', end='\r')
            for bead_val, bead in enumerate(g_var.cg_residues[residue_type][residue]):
                bead_info = g_var.cg_residues[residue_type][residue][bead]
                if g_var.args.box != None and residue_type not in ['PROTEIN', 'OTHER']:
                    cut = check_new_box(g_var.cg_residues[residue_type][residue][bead]['coord'],box_vec.split()[1:4], new_box)
                    if cut:
                        cut_keys.append(residue)
                        break
                if residue_type in ['PROTEIN', 'OTHER'] and bead in g_var.res_top[bead_info['residue_name']]['CONNECT']:
                    if residue != BB_pre_resid and residue != 0:
                        BB_pre_resid = residue
                        con_info = g_var.res_top[bead_info['residue_name']]['CONNECT'][bead]
                        for con_val, connection in enumerate(con_info['dir']):
                            if connection < 0:
                                BB_cur = g_var.cg_residues[residue_type][residue][bead]['coord']
                                if residue+connection in g_var.cg_residues[residue_type]:
                                    BB_pre = g_var.cg_residues[residue_type][residue+connection][con_info['Con_Bd'][con_val]]['coord']
                                    g_var.cg_residues[residue_type][residue][bead]['coord'] = brute_mic(BB_pre, BB_cur)

                if bead_val != 0:
                    g_var.cg_residues[residue_type][residue][bead]['coord'] = brute_mic(g_var.cg_residues[residue_type][residue][bead_prev]['coord'],
                                                                                    g_var.cg_residues[residue_type][residue][bead]['coord'])
                bead_prev=bead
                if g_var.args.box != None:
                    g_var.cg_residues[residue_type][residue][bead]['coord'] = g_var.cg_residues[residue_type][residue][bead]['coord']-box_shift
        for key in cut_keys:
            g_var.cg_residues[residue_type].pop(key)
    print('{:<100}'.format(''), end='\r')

def swap(atom, residue, resid):
    if residue in g_var.swap_dict:
        for key, value in g_var.swap_dict[residue].items():
            break
        if 'ALL' in g_var.swap_dict[residue][key]['resid'] or resid in g_var.swap_dict[residue][key]['resid']:
            if atom in g_var.swap_dict[residue][key]:
                atom = g_var.swap_dict[residue][key][atom].upper()
            residue = key.split(':')[1].upper()
    return atom, residue


##################################################################  User supplied protein ##############

def read_in_atomistic(protein):
#### reset location and check if pdb exists  
    os.chdir(g_var.start_dir)
    if not os.path.exists(protein):
        sys.exit('cannot find atomistic protein : '+protein)
#### read in atomistic fragments into dictionary residue_list[0]=x,y,z,atom_name    
    atomistic_protein_input={}
    if g_var.input_directory in protein:
        chain_count=g_var.chain_count
    else:
        chain_count = 0
#### read in pdb
    new_chain= False
    with open(protein, 'r') as pdb_input:
        pdb_lines_atoms = filter_input(pdb_input.readlines(), False)
        atomistic_protein_input[chain_count]={}
        for line_nr, line_sep in enumerate(pdb_lines_atoms):
            if line_sep['residue_name'] in g_var.alt_res_name:
                line_sep['residue_name'] = g_var.alt_res_name[line_sep['residue_name']]
            if not gen.is_hydrogen(line_sep['atom_name']) or line_sep['residue_name'] in g_var.mod_residues:
                if line_sep['residue_name'] in g_var.p_residues:
                    if 'line_sep_prev' not in locals():
                        line_sep_prev = line_sep.copy()
                        line_sep_prev['residue_id'] = 'X'
                    #### sorts out wrong atoms in terminal residues
                    if line_sep['atom_name'] in ['OT', 'O1', 'O2']:
                        line_sep['atom_name']='O'
                #### makes C_terminal connecting atom variable  
                    if 'prev_atom_coord' in locals():
                        line_sep['x'],line_sep['y'],line_sep['z'] = brute_mic(prev_atom_coord, [line_sep['x'],line_sep['y'],line_sep['z']])

                    if line_sep['residue_id'] !=  line_sep_prev['residue_id']:
                        dir_list = []
                        for directions in g_var.res_top[line_sep['residue_name']]['CONNECT']['atoms'].values():
                            dir_list.append(directions)
                        line_sep_prev = line_sep.copy()
                        if not np.any(np.array(dir_list) < 0 ) or new_chain:
                            if len(atomistic_protein_input[chain_count]) != 0 :
                                chain_count+=1
                                atomistic_protein_input[chain_count]={}
                    if line_sep['atom_name'] in g_var.res_top[line_sep['residue_name']]['CONNECT']['atoms']:
                        if g_var.res_top[line_sep['residue_name']]['CONNECT']['atoms'][line_sep['atom_name']] > 0:
                            C_ter=[line_sep['x'],line_sep['y'],line_sep['z']]
                            C_resid=line_sep['residue_id']
                        elif 'C_ter' in locals():
                            N_resid=line_sep['residue_id']
                            N_ter=[line_sep['x'],line_sep['y'],line_sep['z']]
                            dist=gen.calculate_distance(N_ter, C_ter)
                            if C_resid != N_resid and dist > 3.5:
                                del N_ter, C_ter
                                chain_count+=1
                                atomistic_protein_input[chain_count]={} ### new chain key

                    
                    prev_atom_coord = [line_sep['x'],line_sep['y'],line_sep['z']]
                    if line_sep['residue_id'] not in atomistic_protein_input[chain_count]:  ## if protein does not exist add to dict
                        atomistic_protein_input[chain_count][line_sep['residue_id']]={}
                #### adds atom to dictionary, every atom is given a initial mass of zero 
                    atomistic_protein_input[chain_count][line_sep['residue_id']][line_sep['atom_number']]={'coord':np.array([line_sep['x'],line_sep['y'],line_sep['z']]),'atom':line_sep['atom_name'], 'res_type':line_sep['residue_name'],'frag_mass':0, 'resid':line_sep['residue_id']}
                #### if atom is in the backbone list then its mass is updated to the correct one
                    if line_sep['atom_name'] in g_var.res_top[line_sep['residue_name']]['ATOMS']:
                        if line_sep['atom_name'] in line_sep['atom_name'] in g_var.res_top[line_sep['residue_name']]['atom_masses']:
                            atomistic_protein_input[chain_count][line_sep['residue_id']][line_sep['atom_number']]['frag_mass']=g_var.res_top[line_sep['residue_name']]['atom_masses'][line_sep['atom_name']]    
    return atomistic_protein_input, chain_count+1    

def duplicate_chain(test=False):
    if len(g_var.args.d) != 0:
        if not test:
            print('{:<100}'.format(''), end='\r')
            print('Now duplicating the supplied chains')
        for ch_d in g_var.args.d:
            duplicate = [int(x) for x in ch_d.split(':')]
            if len(duplicate) == 2 and duplicate[0] in g_var.atomistic_protein_input_raw:
                if not test:
                    print('Using '+str(duplicate[1])+' copies of the atomistic chain '+str(duplicate[0]))
                for chain_duplication in range(duplicate[1]-1):
                    g_var.atomistic_protein_input_raw[g_var.chain_count]=copy.deepcopy(g_var.atomistic_protein_input_raw[duplicate[0]])
                    g_var.chain_count+=1
            else:
                sys.exit('your atomistic chain duplication input is incorrrect')   



#!/usr/bin/env python3

import os, sys
import gen, g_var


def write_posre_file(residue, posres):
    with open(residue+'_posre.itp', 'w') as posre_output:
        posre_output.write('; position restraints for '+g_var.args.posre+'\n[ position_restraints ]\n;  i    funct       fcx        fcy        fcz\n')
        for atom in posres[g_var.args.posre]:
            posre_output.write(atom)
        print('Created posres file for '+g_var.args.posre)

def append_ifdef(frag_location):
    with open(frag_location, 'a') as itp_input:
        itp_input.write('\n; Include heavy atom position restraint file\n#ifdef NP\n#include "'+g_var.args.posre+'_posre.itp"\n#endif\n')
    print('Added ifdef flag to '+g_var.args.posre)

def read_itp(filename_itp, footer=False):
    posre = {}
    molecule ={}
    mol_type = False
    atoms=False
    if os.path.exists(filename_itp):
        with open(filename_itp, 'r') as itp_input:
            for line in itp_input.readlines():
                if not (line.isspace() or len(line) == 0) and not line.startswith(';'):
                    line_sep = line.split()
                    if line.startswith('[') and gen.strip_header(line) == 'moleculetype':
                        mol_type = True
                    elif line.startswith('[') and gen.strip_header(line) == 'atoms':
                        atoms, mol_type = True, False
                    elif line.startswith('['):    
                        atoms = False
                        mol_type = False  
                    elif mol_type:
                        molecule[line_sep[0]] = []
                        posre[line_sep[0]] = []
                        mol = line_sep[0]
                    elif atoms:
                        molecule[mol].append(line_sep[4])
                        if len(line.split()) >= 8:
                            if int(float(line.split()[7])) > 1:
                                posre[mol].append('   {0:5}{1:3}{2:11}{3:11}{4:11}\n'.format(line.split()[0],1,50,50,50))
                    if '#ifdef NP' in line:
                        footer=True
        return molecule, posre, footer
    else:
        sys.exit('Cannot find itp file: '+filename_itp)
    

def check_frag_file(directory, molecule):
    exists = []
    wrong = []
    for mol, beads in molecule.items():
        beads.sort()
        bead_list = []
        if os.path.exists(directory+mol+'/'+mol+'.pdb'):
            exists.append(mol)
            with open(directory+mol+'/'+mol+'.pdb', 'r') as pdb_input:
                for line_nr, line in enumerate(pdb_input.readlines()):
                    if line.startswith('['):
                        bead_list.append(gen.strip_header(line))
                bead_list.sort()

            if bead_list == beads:
                print(mol, 'Correct')
            else:
                l1=mol+' Incorrect: '+directory+mol+'/'+mol+'.pdb'
                l2='\nMartini '+mol+': '+', '.join(map(str, beads))
                l3='\nCG2AT2  '+mol+': '+', '.join(map(str, bead_list))
                wrong.append([l1+l2+l3])
    for residue in wrong:
        print(residue)
    return exists

def check_fragments_missing_from_itp(directory, molecule):
    print('\n-------------------')
    print('Fragment files missing from itp\n')
    for filename in os.listdir(directory):
        if filename not in molecule:
            print(filename)

def check_fragments_to_add(directory, molecule, exists):
    print('\n-------------------')
    print('Fragment files to add\n')   
    count = 0 
    for mol, beads in molecule.items():
        if not os.path.exists(directory+mol+'/'+mol+'.pdb'):
            count+=1
            print('Fragment entry required for: ',mol)
    print('Currently missing: ',count)

def add_posres_file():
    print()
    frag_location=gen.fragment_location(g_var.args.posre)[:-4]
    molecule, posre, footer = read_itp(frag_location+'.itp')
    write_posre_file(frag_location, posre)
    if not footer:
        append_ifdef(frag_location+'.itp')
    sys.exit()

def compare_forcefield_to_database():
    molecule, posre, footer = read_itp(g_var.args.compare)
    for directory in g_var.np_directories:
        print('\nChecking: ', directory[0],'\n')
        exists = check_frag_file(directory[0], molecule)
        check_fragments_missing_from_itp(directory[0], molecule)
        check_fragments_to_add(directory[0], molecule, exists)
    sys.exit()

